import polars as pl
import pytest
from polars.testing import assert_frame_equal
from sqlalchemy import Engine

from matchbox.client.dags import DAG
from matchbox.client.results import ModelResults, ResolvedMatches
from matchbox.client.sources import Source
from matchbox.common.arrow import SCHEMA_QUERY_WITH_LEAVES
from matchbox.common.factories.sources import source_factory, source_from_tuple


class TestModelResults:
    """Test ModelResult objects."""

    def test_duplicate_removal(self) -> None:
        """Removes redundant pairs, keeping lowest probability."""
        simple_duplicate = pl.DataFrame(
            [
                {"left_id": 4, "right_id": 5, "probability": 50},
                {"left_id": 4, "right_id": 5, "probability": 100},
            ]
        )

        assert_frame_equal(
            ModelResults(probabilities=simple_duplicate).probabilities,
            simple_duplicate.tail(1),
            check_row_order=False,
            check_column_order=False,
            check_dtypes=False,
        )

        symmetric_duplicate = pl.DataFrame(
            [
                {"left_id": 5, "right_id": 4, "probability": 50},
                {"left_id": 4, "right_id": 5, "probability": 100},
            ]
        )

        assert_frame_equal(
            ModelResults(probabilities=symmetric_duplicate).probabilities,
            symmetric_duplicate.tail(1),
            check_row_order=False,
            check_column_order=False,
            check_dtypes=False,
        )

        no_duplicates = pl.DataFrame(
            [
                {"left_id": 4, "right_id": 6, "probability": 50},
                {"left_id": 4, "right_id": 5, "probability": 100},
            ]
        )

        assert_frame_equal(
            ModelResults(probabilities=no_duplicates).probabilities,
            no_duplicates,
            check_row_order=False,
            check_column_order=False,
            check_dtypes=False,
        )

    def test_clusters_and_root_leaf(self) -> None:
        """From a results object, we can derive clusters at various levels."""
        # Prepare dummy data and model
        left_root_leaf = pl.DataFrame(
            [
                # Two keys per root
                {"id": 10, "leaf_id": 1},
                {"id": 10, "leaf_id": 1},
                # Two leaves per root (same representation)
                {"id": 20, "leaf_id": 2},
                {"id": 20, "leaf_id": 3},
                # Singleton cluster with two keys
                {"id": 4, "leaf_id": 4},
                {"id": 4, "leaf_id": 4},
            ]
        )

        right_root_leaf = pl.DataFrame(
            # For simplicity, all these are singleton clusters
            [
                {"id": 5, "leaf_id": 5},
                {"id": 6, "leaf_id": 6},
                {"id": 7, "leaf_id": 7},
                {"id": 8, "leaf_id": 8},
            ]
        )

        probabilities = pl.DataFrame(
            [
                # Simple left-right merge
                {"left_id": 4, "right_id": 5, "probability": 100},
                # Dedupe through linking
                {"left_id": 10, "right_id": 6, "probability": 100},
                {"left_id": 10, "right_id": 7, "probability": 100},
            ]
        )

        results = ModelResults(
            probabilities=probabilities,
            left_root_leaf=left_root_leaf,
            right_root_leaf=right_root_leaf,
        )

        # Check two ways of representing clusters
        clusters = results.clusters
        grouped_children = {
            tuple(sorted(group))
            for group in clusters.group_by("parent").agg("child")["child"]
        }
        # Only input IDs referenced in probabilities are present, in the right groups
        assert grouped_children == {(4, 5), (6, 7, 10)}

        root_leaf = results.root_leaf()
        grouped_leaves = {
            tuple(sorted(group))
            for group in root_leaf.group_by("root_id").agg("leaf_id")["leaf_id"]
        }
        # Only single-digits are present, and all of them, in the right groups
        assert grouped_leaves == {(1, 6, 7), (2, 3), (4, 5), (8,)}

        # To check edge cases, look at no probabilities returned
        empty_results = ModelResults(
            probabilities=pl.DataFrame(
                {"left_id": [], "right_id": [], "probability": []}
            ),
            left_root_leaf=left_root_leaf,
            right_root_leaf=right_root_leaf,
        )

        assert len(empty_results.clusters) == 0
        expected_empty_root_leaf = pl.concat(
            [
                left_root_leaf.rename({"id": "root_id"}),
                right_root_leaf.rename({"id": "root_id"}),
            ]
        ).unique()
        assert_frame_equal(
            empty_results.root_leaf(),
            expected_empty_root_leaf,
            check_column_order=False,
            check_row_order=False,
        )

        # The above was only possible because leaf IDs were present in the inputs
        only_prob_results = ModelResults(probabilities=probabilities)
        with pytest.raises(RuntimeError, match="instantiated for validation"):
            only_prob_results.root_leaf()


class TestResolvedMatches:
    """Test ResolvedMatches objects."""

    @pytest.fixture(scope="function")  # warehouse is function-scoped
    def dummy_data(
        self, sqlite_in_memory_warehouse: Engine
    ) -> tuple[Source, Source, pl.DataFrame, pl.DataFrame]:
        """Create foo, bar and associated matches."""
        foo = (
            source_from_tuple(
                name="foo",
                engine=sqlite_in_memory_warehouse,
                data_keys=["1", "2", "2b", "3"],
                data_tuple=(
                    {"field_a": 10},
                    {"field_a": 20},
                    {"field_a": 20},
                    {"field_a": 30},
                ),
            )
            .write_to_location()
            .source
        )
        bar = (
            source_from_tuple(
                name="bar",
                engine=sqlite_in_memory_warehouse,
                data_keys=["a", "b", "c", "d"],
                data_tuple=(
                    {"field_a": "1x", "field_b": "1y"},
                    {"field_a": "2x", "field_b": "2y"},
                    {"field_a": "3x", "field_b": "3y"},
                    {"field_a": "4x", "field_b": "4y"},
                ),
            )
            .write_to_location()
            .source
        )

        # Both foo and bar have a record that's not linked
        # Foo has two keys for one leaf ID
        # Foo and bar have links across; bar also has link within
        foo_query_data = pl.DataFrame(
            [
                {"id": 14, "leaf_id": 1, "key": "1"},
                {"id": 2, "leaf_id": 2, "key": "2"},
                {"id": 2, "leaf_id": 2, "key": "2b"},
                {"id": 356, "leaf_id": 3, "key": "3"},
            ],
            schema=pl.Schema(SCHEMA_QUERY_WITH_LEAVES),
        )

        bar_query_data = pl.DataFrame(
            [
                {"id": 14, "leaf_id": 4, "key": "a"},
                {"id": 356, "leaf_id": 5, "key": "b"},
                {"id": 356, "leaf_id": 6, "key": "c"},
                {"id": 7, "leaf_id": 7, "key": "d"},
            ],
            schema=pl.Schema(SCHEMA_QUERY_WITH_LEAVES),
        )

        return foo, bar, foo_query_data, bar_query_data

    def test_from_dump(
        self, dummy_data: tuple[Source, Source, pl.DataFrame, pl.DataFrame]
    ) -> None:
        """Can initialise ResolvedMatches from concatenated dataframe representation."""
        foo, bar, foo_query_data, bar_query_data = dummy_data
        # These won't be the same sources as above, but we only need them them
        # to have the same name
        dag = DAG("companies")
        dag.source(**source_factory(name="foo").into_dag())
        dag.source(**source_factory(name="bar").into_dag())

        original = ResolvedMatches(
            sources=[foo, bar], query_results=[foo_query_data, bar_query_data]
        )

        new = ResolvedMatches.from_dump(cluster_key_map=original.as_dump(), dag=dag)

        assert new.sources[0] == dag.get_source("foo")
        assert new.sources[1] == dag.get_source("bar")

        assert_frame_equal(
            foo_query_data,
            new.query_results[0],
            check_row_order=False,
            check_column_order=False,
        )

        assert_frame_equal(
            bar_query_data,
            new.query_results[1],
            check_row_order=False,
            check_column_order=False,
        )

    def test_as_lookup(
        self, dummy_data: tuple[Source, Source, pl.DataFrame, pl.DataFrame]
    ) -> None:
        """Lookup can be generated from resolved data."""
        foo, bar, foo_query_data, bar_query_data = dummy_data

        # Because of FULL OUTER JOIN, we expect some nulls, and some explosions
        expected_foo_bar_mapping = pl.DataFrame(
            [
                {"id": 14, "foo_key": "1", "bar_key": "a"},
                {"id": 2, "foo_key": "2", "bar_key": None},
                {"id": 2, "foo_key": "2b", "bar_key": None},
                {"id": 356, "foo_key": "3", "bar_key": "b"},
                {"id": 356, "foo_key": "3", "bar_key": "c"},
                {"id": 7, "foo_key": None, "bar_key": "d"},
            ]
        )

        # When selecting single source, we won't explode
        expected_foo_mapping = pl.DataFrame(
            [
                {"id": 14, "foo_key": "1"},
                {"id": 2, "foo_key": "2"},
                {"id": 2, "foo_key": "2b"},
                {"id": 356, "foo_key": "3"},
            ]
        )

        # Retrieve single table
        foo_mapping = ResolvedMatches(
            sources=[foo], query_results=[foo_query_data]
        ).as_lookup()

        assert_frame_equal(
            foo_mapping,
            expected_foo_mapping,
            check_row_order=False,
            check_column_order=False,
        )

        # Retrieve multiple tables
        foo_bar_mapping = ResolvedMatches(
            sources=[foo, bar], query_results=[foo_query_data, bar_query_data]
        ).as_lookup()

        assert_frame_equal(
            foo_bar_mapping,
            expected_foo_bar_mapping,
            check_row_order=False,
            check_column_order=False,
        )

    def test_as_dump(
        self, dummy_data: tuple[Source, Source, pl.DataFrame, pl.DataFrame]
    ) -> None:
        """Mapping across root, leaf, source and key can be generated."""
        foo, bar, foo_query_data, bar_query_data = dummy_data

        mapping = ResolvedMatches(
            sources=[foo, bar], query_results=[foo_query_data, bar_query_data]
        ).as_dump()

        expected_mapping = pl.DataFrame(
            [
                {"source": "foo", "id": 14, "leaf_id": 1, "key": "1"},
                {"source": "foo", "id": 2, "leaf_id": 2, "key": "2"},
                {"source": "foo", "id": 2, "leaf_id": 2, "key": "2b"},
                {"source": "foo", "id": 356, "leaf_id": 3, "key": "3"},
                {"source": "bar", "id": 14, "leaf_id": 4, "key": "a"},
                {"source": "bar", "id": 356, "leaf_id": 5, "key": "b"},
                {"source": "bar", "id": 356, "leaf_id": 6, "key": "c"},
                {"source": "bar", "id": 7, "leaf_id": 7, "key": "d"},
            ]
        )

        assert_frame_equal(
            mapping, expected_mapping, check_row_order=False, check_column_order=False
        )

    def test_as_leaf_sets(
        self, dummy_data: tuple[Source, Source, pl.DataFrame, pl.DataFrame]
    ) -> None:
        """Can generate grouping of lead IDs."""
        foo, bar, foo_query_data, bar_query_data = dummy_data
        leaf_sets = ResolvedMatches(
            sources=[foo, bar], query_results=[foo_query_data, bar_query_data]
        ).as_leaf_sets()

        assert sorted(leaf_sets) == sorted([[1, 4], [2], [3, 5, 6], [7]])

    def test_view_cluster(
        self, dummy_data: tuple[Source, Source, pl.DataFrame, pl.DataFrame]
    ) -> None:
        """Single cluster can be viewed with source data."""
        foo, bar, foo_query_data, bar_query_data = dummy_data

        cluster = ResolvedMatches(
            sources=[foo, bar], query_results=[foo_query_data, bar_query_data]
        ).view_cluster(356)

        # Expanded representation
        expected_cluster = pl.DataFrame(
            {
                "foo_key": ["3", None, None],
                "foo_field_a": [30, None, None],
                "bar_key": [None, "b", "c"],
                "bar_field_a": [None, "2x", "3x"],
                "bar_field_b": [None, "2y", "3y"],
            }
        )

        assert_frame_equal(
            cluster, expected_cluster, check_row_order=False, check_column_order=False
        )

        # Compact representation
        cluster_merged = ResolvedMatches(
            sources=[foo, bar], query_results=[foo_query_data, bar_query_data]
        ).view_cluster(356, merge_fields=True)

        # Note: 30 gets cast to a string
        expected_cluster_merged = pl.DataFrame(
            [
                {"foo_key": "3", "bar_key": None, "field_a": "30", "field_b": None},
                {"foo_key": None, "bar_key": "b", "field_a": "2x", "field_b": "2y"},
                {"foo_key": None, "bar_key": "c", "field_a": "3x", "field_b": "3y"},
            ]
        )

        assert_frame_equal(
            cluster_merged,
            expected_cluster_merged,
            check_row_order=False,
            check_column_order=False,
        )

        # View cluster with multiple keys per leaf, but only one source
        cluster_convergent = ResolvedMatches(
            sources=[foo, bar], query_results=[foo_query_data, bar_query_data]
        ).view_cluster(2)
        expected_cluster_convergent = pl.DataFrame(
            [
                # No columns from "bar"
                {"foo_key": "2", "foo_field_a": 20},
                {"foo_key": "2b", "foo_field_a": 20},
            ]
        )

        assert_frame_equal(
            cluster_convergent,
            expected_cluster_convergent,
            check_row_order=False,
            check_column_order=False,
        )

    def test_merge(
        self, dummy_data: tuple[Source, Source, pl.DataFrame, pl.DataFrame]
    ) -> None:
        """Can merge two instances of resolved matches."""
        # Define first resolved matches
        foo, bar, foo_query_data, bar_query_data = dummy_data

        resolved_one = ResolvedMatches(
            sources=[foo, bar], query_results=[foo_query_data, bar_query_data]
        )

        # Define alternative resolved matches
        alt_foo_query_data = pl.DataFrame(
            [
                {"id": 12, "leaf_id": 1, "key": "1"},
                {"id": 12, "leaf_id": 2, "key": "2"},
                {"id": 12, "leaf_id": 2, "key": "2b"},
                {"id": 35, "leaf_id": 3, "key": "3"},
            ],
            schema=pl.Schema(SCHEMA_QUERY_WITH_LEAVES),
        )

        alt_bar_query_data = pl.DataFrame(
            [
                {"id": 4, "leaf_id": 4, "key": "a"},
                {"id": 35, "leaf_id": 5, "key": "b"},
                {"id": 6, "leaf_id": 6, "key": "c"},
            ],
            schema=pl.Schema(SCHEMA_QUERY_WITH_LEAVES),
        )

        resolved_two = ResolvedMatches(
            sources=[foo, bar], query_results=[alt_foo_query_data, alt_bar_query_data]
        )

        # Merge the two
        merged_resolved = resolved_one.merge(resolved_two)
        new_clusters = merged_resolved.as_dump().group_by("id").agg("leaf_id")
        # The sources are unchanged
        assert merged_resolved.sources == [foo, bar]
        # All cluster IDs now have negative integers
        assert (new_clusters["id"] < 0).all()
        # New clusters contain leaves we expect
        leaf_sets = []
        for ls in new_clusters["leaf_id"].to_list():
            # We remove duplicates created by multiple keys for same leaf
            leaf_sets.append(sorted(set(ls)))

        assert sorted(leaf_sets) == [[1, 2, 4], [3, 5, 6], [7]]



"""Evaluation PostgreSQL mixin for Matchbox server."""

from datetime import UTC, datetime, timedelta
from itertools import chain
from typing import TYPE_CHECKING, Any

import numpy as np
import polars as pl
import pyarrow as pa
from pyarrow import Table
from sqlalchemy import BIGINT, func, select
from sqlalchemy.orm import Session

from matchbox.common.arrow import (
    SCHEMA_CLUSTER_EXPANSION,
    SCHEMA_EVAL_SAMPLES,
    SCHEMA_JUDGEMENTS,
)
from matchbox.common.db import sql_to_df
from matchbox.common.dtos import ModelResolutionPath
from matchbox.common.eval import Judgement as CommonJudgement
from matchbox.common.exceptions import (
    MatchboxTooManySamplesRequested,
    MatchboxUserNotFoundError,
)
from matchbox.common.transform import hash_cluster_leaves
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.orm import (
    Clusters,
    ClusterSourceKey,
    Contains,
    EvalJudgements,
    Probabilities,
    Resolutions,
    SourceConfigs,
    Users,
)
from matchbox.server.postgresql.utils.db import (
    compile_sql,
    ingest_to_temporary_table,
)

if TYPE_CHECKING:
    from pyarrow import Table as ArrowTable
else:
    ArrowTable = Any


class MatchboxPostgresEvaluationMixin:
    """Evaluation mixin for the PostgreSQL adapter for Matchbox."""

    def insert_judgement(self, judgement: CommonJudgement) -> None:  # noqa: D102
        def get_or_create_cluster(leaves: list[int], session: Session) -> int:
            # Compute hash corresponding to set of source clusters (leaves)
            leaf_hashes = [
                session.scalar(
                    select(Clusters.cluster_hash).where(Clusters.cluster_id == leaf_id)
                )
                for leaf_id in leaves
            ]
            cluster_hash = hash_cluster_leaves(leaf_hashes)

            # Add new cluster to session, if it doesn't exist already
            if not (
                cluster_id := session.scalar(
                    select(Clusters.cluster_id).where(
                        Clusters.cluster_hash == cluster_hash
                    )
                )
            ):
                new_c = Clusters(cluster_hash=cluster_hash)
                session.add(new_c)
                session.flush()
                cluster_id = new_c.cluster_id
                for leaf_id in leaves:
                    session.add(Contains(root=cluster_id, leaf=leaf_id))

            return cluster_id

        # Check that all referenced leaf IDs exist
        ids = list(chain(*judgement.endorsed)) + judgement.shown
        self.validate_ids(ids)

        # Check that the user exists
        with MBDB.get_session() as session:
            user = session.scalar(
                select(Users).where(Users.name == judgement.user_name)
            )
            if not user:
                raise MatchboxUserNotFoundError(
                    f"User '{judgement.user_name}' not found"
                )

        with MBDB.get_session() as session:
            shown_id = get_or_create_cluster(judgement.shown, session)
            session.commit()

            for leaves in judgement.endorsed:
                # If cluster with this hash does not exist, create it.
                # Note that only endorsed clusters might be new. The cluster shown to
                # the user is guaranteed to exist in the backend; we have checked above.

                session.add(
                    EvalJudgements(
                        user_id=user.user_id,
                        tag=judgement.tag,
                        shown_cluster_id=shown_id,
                        endorsed_cluster_id=get_or_create_cluster(leaves, session),
                        timestamp=datetime.now(UTC),
                    )
                )
                session.commit()

    def get_judgements(self, tag: str | None = None) -> tuple[Table, Table]:  # noqa: D102
        def _cast_tables(
            judgements: pl.DataFrame, cluster_expansion: pl.DataFrame
        ) -> tuple[pa.Table, pa.Table]:
            """Cast judgement tables to conform to data transfer schema."""
            judgements = judgements.cast(pl.Schema(SCHEMA_JUDGEMENTS))
            cluster_expansion = cluster_expansion.cast(
                pl.Schema(SCHEMA_CLUSTER_EXPANSION)
            )

            return (
                judgements.to_arrow().cast(SCHEMA_JUDGEMENTS),
                cluster_expansion.to_arrow().cast(SCHEMA_CLUSTER_EXPANSION),
            )

        judgements_stmt = select(
            Users.name.label("user_name"),
            EvalJudgements.endorsed_cluster_id.label("endorsed"),
            EvalJudgements.shown_cluster_id.label("shown"),
        ).join(Users)

        if tag:
            judgements_stmt = judgements_stmt.where(EvalJudgements.tag == tag)

        with MBDB.get_adbc_connection() as conn:
            judgements = sql_to_df(
                stmt=compile_sql(judgements_stmt),
                connection=conn.dbapi_connection,
                return_type="polars",
            )

        if not len(judgements):
            cluster_expansion = pl.DataFrame(schema=pl.Schema(SCHEMA_CLUSTER_EXPANSION))
            return _cast_tables(judgements, cluster_expansion)

        shown_clusters = set(judgements["shown"].to_list())
        endorsed_clusters = set(judgements["endorsed"].to_list())
        referenced_clusters = Table.from_pydict(
            {"root": list(shown_clusters | endorsed_clusters)}
        )

        with ingest_to_temporary_table(
            table_name="judgements",
            schema_name="mb",
            column_types={
                "root": BIGINT(),
            },
            data=referenced_clusters,
        ) as temp_table:
            cluster_expansion_stmt = (
                select(temp_table.c.root, func.array_agg(Contains.leaf).label("leaves"))
                .select_from(temp_table)
                .join(Contains, Contains.root == temp_table.c.root)
                .group_by(temp_table.c.root)
            )

            with MBDB.get_adbc_connection() as conn:
                cluster_expansion = sql_to_df(
                    stmt=compile_sql(cluster_expansion_stmt),
                    connection=conn.dbapi_connection,
                    return_type="polars",
                )

        return _cast_tables(judgements, cluster_expansion)

    def sample_for_eval(  # noqa: D102
        self, n: int, path: ModelResolutionPath, user_name: str
    ) -> ArrowTable:
        """Sample some clusters from a resolution."""
        # Not currently checking validity of the user_name
        # If the user ID does not exist, the exclusion by previous judgements breaks
        if n > 100:
            # This reasonable assumption means simple "IS IN" function later is fine
            raise MatchboxTooManySamplesRequested(
                "Can only sample 100 entries at a time."
            )

        with MBDB.get_session() as session:
            # Get user
            user = session.scalar(select(Users).where(Users.name == user_name))
            if not user:
                raise MatchboxUserNotFoundError(f"User '{user_name}' not found")

            # Use ORM to get resolution metadata
            resolution_orm = Resolutions.from_path(path=path, session=session)
            resolution_id = resolution_orm.resolution_id
            truth = resolution_orm.truth

        # Get a list of cluster IDs and features for this resolution and user
        user_judgements = (
            select(EvalJudgements)
            .where(EvalJudgements.user_id == user.user_id)
            .subquery()
        )
        cluster_features_stmt = (
            select(
                Probabilities.cluster_id,
                # We expect only one probability per cluster within one resolution
                func.max(Probabilities.probability).label("probability"),
                func.max(user_judgements.c.timestamp).label("latest_ts"),
            )
            .join(
                user_judgements,
                Probabilities.cluster_id == user_judgements.c.shown_cluster_id,
                isouter=True,
            )
            .where(
                Probabilities.resolution_id == resolution_id,
            )
            .group_by(Probabilities.cluster_id)
        )

        with MBDB.get_adbc_connection() as conn:
            cluster_features = sql_to_df(
                stmt=compile_sql(cluster_features_stmt),
                connection=conn.dbapi_connection,
                return_type="polars",
            )

        # Exclude clusters recently judged by this user
        to_sample = cluster_features.filter(
            (pl.col("latest_ts") < datetime.now(UTC) - timedelta(days=365))
            | (pl.col("latest_ts").is_null())
        )

        # Return early if nothing to sample from
        if not len(to_sample):
            return pl.DataFrame(schema=pl.Schema(SCHEMA_EVAL_SAMPLES)).to_arrow()

        # Sample proportionally to distance from the truth, and get 1D array
        distances = np.abs(to_sample.select("probability").to_numpy() - truth)[:, 0]
        # Add small noise to avoid division by 0 if all distances are 0
        unnormalised_probs = distances + 0.001
        probs = unnormalised_probs / unnormalised_probs.sum()

        # With fewer clusters than requested, return all
        if to_sample.shape[0] <= n:
            sampled_cluster_ids = to_sample.select("cluster_id").to_series().to_list()
        else:
            indices = np.random.choice(
                to_sample.shape[0], size=n, p=probs, replace=False
            )
            sampled_cluster_ids = (
                to_sample[indices].select("cluster_id").to_series().to_list()
            )

        # Get all info we need for the cluster IDs we've sampled, i.e.:
        # source cluster IDs, keys and source resolutions
        with MBDB.get_adbc_connection() as conn:
            source_clusters = (
                select(Contains.root, Contains.leaf)
                .where(Contains.root.in_(sampled_cluster_ids))
                .subquery()
            )

            # The same leaf can be reused to represent rows across different sources
            # We only want to retrieve info for sources upstream of resolution
            source_resolution_ids = [
                res.resolution_id
                for res in resolution_orm.ancestors
                if res.type == "source"
            ]
            source_resolutions = (
                select(Resolutions.name, Resolutions.resolution_id)
                .where(Resolutions.resolution_id.in_(source_resolution_ids))
                .subquery()
            )

            enrich_stmt = (
                select(
                    source_clusters.c.root,
                    source_clusters.c.leaf,
                    ClusterSourceKey.key,
                    source_resolutions.c.name.label("source"),
                )
                .select_from(source_clusters)
                .join(
                    ClusterSourceKey,
                    ClusterSourceKey.cluster_id == source_clusters.c.leaf,
                )
                .join(
                    SourceConfigs,
                    SourceConfigs.source_config_id == ClusterSourceKey.source_config_id,
                )
                .join(
                    source_resolutions,
                    source_resolutions.c.resolution_id == SourceConfigs.resolution_id,
                )
            )

            final_samples = sql_to_df(
                stmt=compile_sql(enrich_stmt),
                connection=conn.dbapi_connection,
                return_type="polars",
            )
            return final_samples.cast(pl.Schema(SCHEMA_EVAL_SAMPLES)).to_arrow()



"""A linking methodology that applies different weights to field comparisons."""

import duckdb
import polars as pl
from pydantic import BaseModel, Field, field_validator

from matchbox.client.models import comparison
from matchbox.client.models.linkers.base import Linker, LinkerSettings


class WeightedComparison(BaseModel):
    """A valid comparison and a weight to give it."""

    comparison: str = Field(
        description="""
            A valid ON clause to compare fields between the left and 
            the right data.

            Use left.field and right.field to refer to fields in the 
            respective sources.

            For example:

            "left.company_name = right.company_name"
        """
    )
    weight: float = Field(
        description="""
            A weight to give this comparison. Use 1 for all comparisons to give
            uniform weight to each.
        """
    )

    @field_validator("comparison")
    @classmethod
    def validate_comparison(cls, v: str) -> str:
        """Validate the comparison string."""
        comp_val = comparison(v)
        return comp_val


class WeightedDeterministicSettings(LinkerSettings):
    """A data class to enforce the Weighted linker's settings dictionary shape.

    Example:
        >>> {
        ...     left_id: "hash",
        ...     right_id: "hash",
        ...     weighted_comparisons: [
        ...         ("l.company_name = r.company_name", 0.7),
        ...         ("l.postcode = r.postcode", 0.7),
        ...         ("l.company_id = r.company_id", 1),
        ...     ],
        ...     threshold: 0.8,
        ... }
    """

    weighted_comparisons: list[WeightedComparison] = Field(
        description="A list of tuples in the form of a comparison, and a weight."
    )
    threshold: float = Field(
        description="""
            The probability above which matches will be kept. 
            
            Inclusive, so a value of 1 will keep only exact matches across all 
            comparisons.
        """,
        ge=0,
        le=1,
    )


class WeightedDeterministicLinker(Linker):
    """A deterministic linker that applies different weights to field comparisons."""

    settings: WeightedDeterministicSettings

    _id_dtype_l: pl.DataType
    _id_dtype_r: pl.DataType

    def prepare(self, left: pl.DataFrame, right: pl.DataFrame) -> None:
        """Prepare the linker for linking."""
        pass

    def link(self, left: pl.DataFrame, right: pl.DataFrame) -> pl.DataFrame:
        """Link the left and right dataframes."""
        self._id_dtype_l = left[self.settings.left_id].dtype
        self._id_dtype_r = right[self.settings.right_id].dtype

        # Used below but ruff can't detect
        left_df = left.clone()  # noqa: F841
        right_df = right.clone()  # noqa: F841

        match_subquery = []
        weights = []

        for weighted_comparison in self.settings.weighted_comparisons:
            match_subquery.append(
                f"""
                    select distinct on (list_sort([raw.left_id, raw.right_id]))
                        raw.left_id,
                        raw.right_id,
                        1.0 * {weighted_comparison.weight} as probability
                    from (
                        select
                            l.{self.settings.left_id} as left_id,
                            r.{self.settings.right_id} as right_id,
                        from
                            left_df l
                        inner join right_df r on
                            {weighted_comparison.comparison}
                    ) raw
                """
            )
            weights.append(weighted_comparison.weight)

        match_subquery = " union all ".join(match_subquery)
        total_weight = sum(weights)

        sql = f"""
            select
                matches.left_id,
                matches.right_id,
                sum(matches.probability) / {total_weight} as probability
            from
                ({match_subquery}) matches
            group by
                matches.left_id,
                matches.right_id
            having
                sum(matches.probability) / 
                    {total_weight} >= {self.settings.threshold};
        """

        return (
            duckdb.sql(sql)
            .pl()
            .with_columns(
                [
                    pl.col("left_id").cast(self._id_dtype_l),
                    pl.col("right_id").cast(self._id_dtype_r),
                    pl.col("probability").cast(pl.Float32),
                ]
            )
        )



"""A module for defining mixins for the PostgreSQL backend ORM."""

from typing import TypeVar

from sqlalchemy import func

from matchbox.server.postgresql.db import MBDB

T = TypeVar("T")


class CountMixin:
    """A mixin for counting the number of rows in a table."""

    @classmethod
    def count(cls: type[T]) -> int:
        """Counts the number of rows in the table."""
        with MBDB.get_session() as session:
            return session.query(func.count()).select_from(cls).scalar()



"""Test the backend adapter's admin functions."""

from functools import partial

import pyarrow as pa
import pytest
from sqlalchemy import Engine
from test.fixtures.db import BACKENDS

from matchbox.common.datatypes import DataTypes
from matchbox.common.dtos import (
    ModelConfig,
    Resolution,
    ResolutionPath,
    SourceConfig,
    SourceField,
    UploadStage,
)
from matchbox.common.exceptions import (
    MatchboxCollectionAlreadyExists,
    MatchboxCollectionNotFoundError,
    MatchboxResolutionAlreadyExists,
    MatchboxResolutionExistingData,
    MatchboxResolutionNotFoundError,
    MatchboxResolutionUpdateError,
    MatchboxRunNotFoundError,
    MatchboxRunNotWriteable,
)
from matchbox.common.factories.entities import diff_results, query_to_cluster_entities
from matchbox.common.factories.models import model_factory
from matchbox.common.factories.scenarios import setup_scenario
from matchbox.common.factories.sources import SourceTestkit
from matchbox.server.base import MatchboxDBAdapter


@pytest.mark.parametrize("backend", BACKENDS)
@pytest.mark.docker
class TestMatchboxCollectionsBackend:
    @pytest.fixture(autouse=True)
    def setup(
        self, backend_instance: MatchboxDBAdapter, sqla_sqlite_warehouse: Engine
    ) -> None:
        self.backend: MatchboxDBAdapter = backend_instance
        self.scenario = partial(setup_scenario, warehouse=sqla_sqlite_warehouse)

    # Collection management

    def test_collections(self) -> None:
        """Test creating, listing, getting and deleting collections."""
        with self.scenario(self.backend, "bare") as _:
            collections_pre = self.backend.list_collections()
            assert "test_collection" not in collections_pre

            # Create new collection and verify its initial properties
            test_collection_created = self.backend.create_collection("test_collection")

            assert test_collection_created.runs == []  # No versions yet

            collections_post = self.backend.list_collections()
            assert "test_collection" in collections_post

            # Verify duplicate creation is rejected
            with pytest.raises(MatchboxCollectionAlreadyExists):
                self.backend.create_collection("test_collection")

            test_collection = self.backend.get_collection("test_collection")

            assert test_collection == test_collection_created

            self.backend.delete_collection("test_collection", certain=True)

            # Verify collection is properly removed
            with pytest.raises(MatchboxCollectionNotFoundError):
                self.backend.get_collection("test_collection")

            with pytest.raises(MatchboxCollectionNotFoundError):
                self.backend.delete_collection("test_collection", certain=False)

    # Run management

    def test_runs(self) -> None:
        """Test creating, listing, getting and deleting runs."""
        with self.scenario(self.backend, "bare") as _:
            collections_pre = self.backend.list_collections()
            assert "test_collection" not in collections_pre

            # Create parent collection with no runs initially
            test_collection_pre = self.backend.create_collection("test_collection")
            assert len(test_collection_pre.runs) == 0  # No runs yet

            # Create first run and check default properties
            v1 = self.backend.create_run("test_collection")

            assert isinstance(v1.run_id, int)
            assert v1.is_default is False  # New runs aren't default by default
            assert v1.is_mutable is True  # New runs are mutable by default
            assert v1.resolutions == {}  # No resolutions yet

            test_collection_post = self.backend.get_collection("test_collection")
            assert v1.run_id in test_collection_post.runs

            v2 = self.backend.create_run("test_collection")

            with pytest.raises(ValueError, match="mutable"):
                self.backend.set_run_default("test_collection", v1.run_id, True)

            self.backend.set_run_mutable("test_collection", v1.run_id, False)
            self.backend.set_run_default("test_collection", v1.run_id, True)

            # Default run info also available on collection DTO
            collection = self.backend.get_collection("test_collection")
            assert collection.default_run == v1.run_id

            # Setting v2 as default should automatically unset v1 as default
            self.backend.set_run_mutable("test_collection", v2.run_id, False)
            self.backend.set_run_default("test_collection", v2.run_id, True)

            v1 = self.backend.get_run("test_collection", v1.run_id)
            v2 = self.backend.get_run("test_collection", v2.run_id)

            assert v1.is_mutable is False
            assert v1.is_default is False

            assert v2.is_mutable is False
            assert v2.is_default is True

            self.backend.delete_run("test_collection", v1.run_id, certain=True)

            # Verify run is properly removed
            with pytest.raises(MatchboxRunNotFoundError):
                self.backend.get_run("test_collection", v1.run_id)

            with pytest.raises(MatchboxRunNotFoundError):
                self.backend.delete_run("test_collection", v1.run_id, certain=False)

    def test_run_immutable(self) -> None:
        """Nothing in an immutable run can be changed."""
        with self.scenario(self.backend, "dedupe") as dag_testkit:
            source_testkit = dag_testkit.sources["crn"]
            model_testkit = dag_testkit.models["naive_test_crn"]

            self.backend.set_run_mutable(
                collection=dag_testkit.dag.name,
                run_id=dag_testkit.dag.run,
                mutable=False,
            )

            with pytest.raises(MatchboxRunNotWriteable):
                self.backend.create_resolution(
                    path=source_testkit.resolution_path.model_copy(
                        update={"name": "new_source"}
                    ),
                    resolution=source_testkit.fake_run().source.to_resolution(),
                )

            with pytest.raises(MatchboxRunNotWriteable):
                self.backend.delete_resolution(
                    source_testkit.resolution_path, certain=True
                )

            with pytest.raises(MatchboxRunNotWriteable):
                self.backend.insert_source_data(
                    source_testkit.resolution_path, source_testkit.data_hashes
                )

            with pytest.raises(MatchboxRunNotWriteable):
                self.backend.insert_model_data(
                    model_testkit.resolution_path,
                    model_testkit.probabilities.to_arrow(),
                )

    # Resolution management

    def test_get_source(self) -> None:
        """Test querying data from the database."""
        with self.scenario(self.backend, "index") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")

            crn_retrieved = self.backend.get_resolution(crn_testkit.resolution_path)
            assert isinstance(crn_retrieved, Resolution)
            assert crn_testkit.source_config == crn_retrieved.config

            with pytest.raises(MatchboxResolutionNotFoundError):
                self.backend.get_resolution(
                    path=ResolutionPath(collection="collection", run=1, name="foo")
                )

    def test_delete_resolution(self) -> None:
        """
        Tests the deletion of:

        * The resolution from the resolutions table
        * The source configuration attached to the resolution
        * Any models that depended on this model (descendants)
        * All probabilities for all descendants
        """
        with self.scenario(self.backend, "link") as dag_testkit:
            to_delete = dag_testkit.sources["crn"]
            total_sources = len(dag_testkit.sources)
            total_models = len(dag_testkit.models)

            source_configs_pre_delete = self.backend.sources.count()
            sources_pre_delete = self.backend.source_resolutions.count()
            models_pre_delete = self.backend.models.count()
            cluster_count_pre_delete = self.backend.model_clusters.count()
            cluster_assoc_count_pre_delete = self.backend.creates.count()
            proposed_merge_probs_pre_delete = self.backend.proposes.count()

            assert sources_pre_delete == total_sources
            assert models_pre_delete == total_models
            assert cluster_count_pre_delete > 0
            assert cluster_assoc_count_pre_delete > 0
            assert proposed_merge_probs_pre_delete > 0

            # Perform deletion
            self.backend.delete_resolution(to_delete.resolution_path, certain=True)

            source_configs_post_delete = self.backend.sources.count()
            sources_post_delete = self.backend.source_resolutions.count()
            models_post_delete = self.backend.models.count()
            cluster_count_post_delete = self.backend.model_clusters.count()
            cluster_assoc_count_post_delete = self.backend.creates.count()
            proposed_merge_probs_post_delete = self.backend.proposes.count()

            # 1 source, 1 index, (1 deduper + 3 linkers) = 4 models are gone
            assert source_configs_post_delete == source_configs_pre_delete - 1
            assert sources_post_delete == sources_pre_delete - 1
            assert models_post_delete == models_pre_delete - 4

            # We've lost some composite clusters
            assert cluster_count_post_delete < cluster_count_pre_delete

            # Count of propose and create edges has dropped
            assert cluster_assoc_count_post_delete < cluster_assoc_count_pre_delete
            assert proposed_merge_probs_post_delete < proposed_merge_probs_pre_delete

    def test_index_new_source(self) -> None:
        """Test that indexing a new source works."""
        with self.scenario(self.backend, "bare") as dag_testkit:
            crn_testkit: SourceTestkit = dag_testkit.sources.get("crn").fake_run()

            assert self.backend.model_clusters.count() == 0

            self.backend.create_resolution(
                crn_testkit.source.to_resolution(),
                path=crn_testkit.resolution_path,
            )

            # Resolution can't be re-added
            with pytest.raises(MatchboxResolutionAlreadyExists):
                self.backend.create_resolution(
                    crn_testkit.source.to_resolution(),
                    path=crn_testkit.resolution_path,
                )

            # After resolution metadata is present, we can add data
            self.backend.insert_source_data(
                crn_testkit.source.resolution_path, crn_testkit.data_hashes
            )

            # Data can't be re-added
            with pytest.raises(MatchboxResolutionExistingData):
                self.backend.insert_source_data(
                    crn_testkit.source.resolution_path, crn_testkit.data_hashes
                )

            # Resolution marked as complete
            assert (
                self.backend.get_resolution_stage(crn_testkit.source.resolution_path)
                == UploadStage.COMPLETE
            )

            # We can retrieve the resolution
            crn_retrieved = self.backend.get_resolution(
                crn_testkit.source.resolution_path
            )

            assert crn_testkit.source_config == crn_retrieved.config
            assert self.backend.source_clusters.count() == len(crn_testkit.data_hashes)

            assert self.backend.source_clusters.count() == len(crn_testkit.data_hashes)
            assert self.backend.source_resolutions.count() == 1

            # We can update the resolution metadata, including changes in fields
            updated_key_field = SourceField(name="new key", type=DataTypes.STRING)
            updated_index_fields = (
                SourceField(name="new field", type=DataTypes.BOOLEAN),
            )
            updated_config = SourceConfig.model_validate(
                crn_testkit.source.config.model_copy(
                    update={
                        "key_field": updated_key_field,
                        "index_fields": updated_index_fields,
                    }
                )
            )
            updated_resolution = Resolution.model_validate(
                crn_testkit.source.to_resolution().model_copy(
                    update={
                        "description": "updated",
                        "config": updated_config,
                    }
                )
            )
            self.backend.update_resolution(
                updated_resolution, path=crn_testkit.source.resolution_path
            )

            # We cannot update source resolution with different fingerprint
            with pytest.raises(MatchboxResolutionUpdateError, match="fingerprint"):
                self.backend.update_resolution(
                    updated_resolution.model_copy(update={"fingerprint": 123}),
                    path=crn_testkit.source.resolution_path,
                )

            # We cannot update source resolution with a model resolution
            with pytest.raises(MatchboxResolutionUpdateError, match="parents"):
                # Create model with same fingerprint as previous resolution
                valid_fingerprint = crn_testkit.source.to_resolution().fingerprint
                model_resolution = Resolution.model_validate(
                    model_factory()
                    .fake_run()
                    .model.to_resolution()
                    .model_copy(update={"fingerprint": valid_fingerprint})
                )
                self.backend.update_resolution(
                    model_resolution,
                    path=crn_testkit.source.resolution_path,
                )

            # We can retrieve the updated resolution
            crn_retrieved = self.backend.get_resolution(
                crn_testkit.source.resolution_path
            )
            assert crn_retrieved.description == "updated"
            assert crn_retrieved.config.key_field == updated_key_field
            assert crn_retrieved.config.index_fields == updated_index_fields

    def test_index_empty_source(self) -> None:
        """Can insert and retrieve empty source data"""
        with self.scenario(self.backend, "bare") as dag_testkit:
            crn_testkit: SourceTestkit = dag_testkit.sources.get("crn")
            crn_testkit.data_hashes = crn_testkit.data_hashes.slice(length=0)
            crn_testkit.fake_run()
            self.backend.create_resolution(
                crn_testkit.source.to_resolution(),
                path=crn_testkit.resolution_path,
            )

            self.backend.insert_source_data(
                crn_testkit.source.resolution_path, crn_testkit.data_hashes
            )
            # Resolution marked as complete
            assert (
                self.backend.get_resolution_stage(crn_testkit.source.resolution_path)
                == UploadStage.COMPLETE
            )

    def test_index_different_resolution_same_hashes(self) -> None:
        """Test that indexing data with the same hashes but different sources works."""
        with self.scenario(self.backend, "bare") as dag_testkit:
            # Prepare original source
            crn_testkit: SourceTestkit = dag_testkit.sources.get("crn").fake_run()
            # Create new source with same hashes
            dh_testkit: SourceTestkit = dag_testkit.sources.get("dh")
            dh_testkit.data_hashes = crn_testkit.data_hashes
            dh_testkit.fake_run()

            # Add original source
            self.backend.create_resolution(
                crn_testkit.source.to_resolution(), path=crn_testkit.resolution_path
            )
            self.backend.insert_source_data(
                crn_testkit.source.resolution_path, crn_testkit.data_hashes
            )
            # Add different source, with same hashes
            self.backend.create_resolution(
                dh_testkit.source.to_resolution(), path=dh_testkit.resolution_path
            )
            self.backend.insert_source_data(
                dh_testkit.source.resolution_path, crn_testkit.data_hashes
            )
            assert self.backend.source_clusters.count() == len(crn_testkit.data_hashes)
            assert self.backend.source_resolutions.count() == 2

    def test_insert_model(self) -> None:
        """Test that models can be inserted."""
        with self.scenario(self.backend, "index") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            # Test deduper insertion
            models_count = self.backend.models.count()

            dedupe_1_testkit = model_factory(
                name="dedupe_1",
                description="Test deduper 1",
                dag=dag_testkit.dag,
                left_testkit=crn_testkit,
                true_entities=linked.true_entities,
            )
            self.backend.create_resolution(
                resolution=dedupe_1_testkit.fake_run().model.to_resolution(),
                path=dedupe_1_testkit.resolution_path,
            )

            dedupe_2_testkit = model_factory(
                name="dedupe_2",
                description="Test deduper 2",
                dag=dag_testkit.dag,
                left_testkit=dh_testkit,
                true_entities=linked.true_entities,
            )
            self.backend.create_resolution(
                resolution=dedupe_2_testkit.fake_run().model.to_resolution(),
                path=dedupe_2_testkit.resolution_path,
            )

            assert self.backend.models.count() == models_count + 2

            # Test linker insertion
            linker_testkit = model_factory(
                name="link_1",
                description="Test linker 1",
                left_testkit=dedupe_1_testkit,
                right_testkit=dedupe_2_testkit,
                true_entities=linked.true_entities,
            )
            self.backend.create_resolution(
                resolution=linker_testkit.fake_run().model.to_resolution(),
                path=linker_testkit.resolution_path,
            )

            assert self.backend.models.count() == models_count + 3

            # We cannot re-create under the same path
            with pytest.raises(MatchboxResolutionAlreadyExists):
                self.backend.create_resolution(
                    linker_testkit.fake_run().model.to_resolution(),
                    path=linker_testkit.resolution_path,
                )

            assert self.backend.models.count() == models_count + 3

            # Can update model resolution
            old_resolution = linker_testkit.model.to_resolution()
            updated_config = ModelConfig.model_validate(
                old_resolution.config.model_copy(
                    update={
                        "left_query": old_resolution.config.left_query.model_copy(
                            update={"threshold": 99}
                        )
                    }
                )
            )
            updated_resolution = Resolution.model_validate(
                old_resolution.model_copy(
                    update={
                        "description": "updated",
                        "truth": 33,
                        "config": updated_config,
                    }
                )
            )
            self.backend.update_resolution(
                resolution=updated_resolution,
                path=linker_testkit.resolution_path,
            )

            # We can retrieve the updated resolution
            linker_retrieved = self.backend.get_resolution(
                linker_testkit.resolution_path
            )
            assert linker_retrieved.description == "updated"
            assert linker_retrieved.truth == 33
            assert linker_retrieved.config.left_query.threshold == 99

            # We cannot change a model's inputs
            rewired_config = ModelConfig.model_validate(
                old_resolution.config.model_copy(
                    update={
                        "left_query": old_resolution.config.left_query.model_copy(
                            update={"model_resolution": "new_model"}
                        )
                    }
                )
            )
            rewired_resolution = Resolution.model_validate(
                old_resolution.model_copy(update={"config": rewired_config})
            )

            with pytest.raises(MatchboxResolutionUpdateError, match="parents"):
                self.backend.update_resolution(
                    resolution=rewired_resolution,
                    path=linker_testkit.resolution_path,
                )

            # We cannot change model results fingerprint
            with pytest.raises(MatchboxResolutionUpdateError, match="fingerprint"):
                corrupt_resolution = Resolution.model_validate(
                    updated_resolution.model_copy(update={"fingerprint": b"fake"})
                )
                self.backend.update_resolution(
                    resolution=corrupt_resolution,
                    path=linker_testkit.resolution_path,
                )

    def test_model_results_basic(self) -> None:
        """Test that a model's results data can be set and retrieved."""
        with self.scenario(self.backend, "dedupe") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            naive_crn_testkit = dag_testkit.models.get("naive_test_crn")

            # Query returns the same results as the testkit, showing
            # that processing was performed accurately.
            # (that we can query from it implies the resolution was correctly
            # marked as complete)
            res = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=naive_crn_testkit.resolution_path,
            )
            res_clusters = query_to_cluster_entities(
                data=res,
                keys={crn_testkit.name: "key"},
            )

            identical, report = diff_results(
                expected=naive_crn_testkit.entities,
                actual=res_clusters,
            )

            assert identical, report

            # Retrieve
            pre_results = self.backend.get_model_data(
                path=naive_crn_testkit.resolution_path
            )

            assert isinstance(pre_results, pa.Table)
            assert len(pre_results) > 0

            # Validate IDs
            self.backend.validate_ids(ids=pre_results["left_id"].to_pylist())
            self.backend.validate_ids(ids=pre_results["right_id"].to_pylist())

            # Cannot set new results
            with pytest.raises(MatchboxResolutionExistingData):
                self.backend.insert_model_data(
                    path=naive_crn_testkit.resolution_path,
                    results=naive_crn_testkit.probabilities.to_arrow(),
                )

            # Retrieve again
            post_results = self.backend.get_model_data(
                path=naive_crn_testkit.resolution_path
            )

            # Check difference
            assert pre_results == post_results

    def test_model_results_probabilistic(self) -> None:
        """Test that a probabilistic model's results data can be set and retrieved."""
        with self.scenario(self.backend, "probabilistic_dedupe") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            prob_crn_testkit = dag_testkit.models.get("probabilistic_test_crn")

            # Query returns the same results as the testkit, showing
            # that processing was performed accurately
            res = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=prob_crn_testkit.resolution_path,
            )
            res_clusters = query_to_cluster_entities(
                data=res,
                keys={crn_testkit.name: "key"},
            )

            identical, report = diff_results(
                expected=prob_crn_testkit.entities,
                actual=res_clusters,
            )
            assert identical, report

            # Retrieve
            pre_results = self.backend.get_model_data(
                path=prob_crn_testkit.resolution_path
            )

            assert isinstance(pre_results, pa.Table)
            assert len(pre_results) > 0

            # Validate IDs
            self.backend.validate_ids(ids=pre_results["left_id"].to_pylist())
            self.backend.validate_ids(ids=pre_results["right_id"].to_pylist())

    def test_model_results_empty(self) -> None:
        """Can insert and retrieve empty model results"""
        with self.scenario(self.backend, "index") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            model_testkit = model_factory(
                model_type="deduper", dag=crn_testkit.source.dag
            )
            model_testkit.probabilities = model_testkit.probabilities.head(0)
            model_testkit.fake_run()

            self.backend.create_resolution(
                model_testkit.model.to_resolution(), path=model_testkit.resolution_path
            )

            self.backend.insert_model_data(
                path=model_testkit.model.resolution_path,
                results=model_testkit.model.results.probabilities.to_arrow(),
            )

            # Querying from deduper with no results is the same as querying from source
            # (That we can query also implies that resolution marked as complete)
            source_query = self.backend.query(source=crn_testkit.resolution_path)
            dedupe_query = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=model_testkit.resolution_path,
            )

            assert source_query == dedupe_query

    def test_model_results_shared_clusters(self) -> None:
        """Test that model results data can be inserted when clusters are shared."""
        with self.scenario(self.backend, "convergent_partial") as dag_testkit:
            for model_testkit in dag_testkit.models.values():
                self.backend.create_resolution(
                    path=model_testkit.resolution_path,
                    resolution=model_testkit.fake_run().model.to_resolution(),
                )
                self.backend.insert_model_data(
                    path=model_testkit.resolution_path,
                    results=model_testkit.probabilities.to_arrow(),
                )



"""Matchbox server.

Includes the API, and database adapters for various backends.
"""

from matchbox.server.base import (
    MatchboxDBAdapter,
    MatchboxServerSettings,
    initialise_matchbox,
)

__all__ = ["MatchboxDBAdapter", "MatchboxServerSettings"]

initialise_matchbox()



"""Objects to define a DAG which indexes, deduplicates and links data."""

import json
import tempfile
from enum import StrEnum
from pathlib import Path
from typing import Any, Self, TypeAlias

import polars as pl
from platformdirs import user_cache_path

from matchbox.client import _handler
from matchbox.client._settings import settings
from matchbox.client.locations import Location
from matchbox.client.models import Model
from matchbox.client.queries import Query
from matchbox.client.results import ResolvedMatches
from matchbox.client.sources import Source
from matchbox.common.dtos import (
    Collection,
    CollectionName,
    ModelResolutionName,
    Resolution,
    ResolutionName,
    ResolutionPath,
    ResolutionType,
    Run,
    RunID,
    SourceResolutionName,
)
from matchbox.common.exceptions import (
    MatchboxCollectionNotFoundError,
    MatchboxResolutionNotFoundError,
)
from matchbox.common.logging import log_mem_usage, logger, profile_time
from matchbox.common.transform import threshold_float_to_int, threshold_int_to_float


class DAGNodeExecutionStatus(StrEnum):
    """Enumeration of node execution statuses."""

    SKIPPED = "skipped"
    DONE = "done"
    DOING = "doing"


DAGExecutionStatus: TypeAlias = dict[str, DAGNodeExecutionStatus]

CACHE_DIR = user_cache_path("matchbox")


class DAG:
    """Self-sufficient pipeline of indexing, deduping and linking steps."""

    def __init__(self, name: str) -> None:
        """Initialises empty DAG."""
        self.name: CollectionName = CollectionName(name)
        self._run: RunID | None = None
        self.nodes: dict[ResolutionName, Source | Model] = {}
        self.graph: dict[ResolutionName, list[ResolutionName]] = {}

        CACHE_DIR.mkdir(parents=True, exist_ok=True)
        self._cache_dir = tempfile.TemporaryDirectory(dir=str(CACHE_DIR))
        self.cache_path = Path(self._cache_dir.name)

    def _check_dag(self, dag: Self) -> None:
        """Check that the given DAG is the same as this one."""
        if self != dag:
            raise ValueError("Cannot mix DAGs")

    def _add_step(self, step: Source | Model) -> None:
        """Validate and add sources and models to DAG."""
        self._check_dag(step.dag)

        # We allow overwriting nodes, because:
        # - One might want to iterate on a local DAG without starting over
        # - One might need to load a pending run in a script which defines the same
        #   DAG, to re-run a specific node.
        #
        # We only mandate that a model's direct inputs not be changed. This ensures
        # that `self.graph` is not altered. It does not guarantee that the new node
        # works well with the rest of the DAG, which must be verified by the user.
        # For example, you could remove source fields that are needed downstream, or
        # you could query sources that are not available to a point of truth.
        # These issues are not checked when adding a node for the first time either.
        if step.name in self.nodes:
            if step.config.parents != self.nodes[step.name].config.parents:
                raise ValueError("Cannot re-assign name to model with different inputs")
            logger.info(f"Overwriting node '{step.name}'.")

        if isinstance(step, Model):
            self._check_dag(step.left_query.dag)
            if step.right_query:
                self._check_dag(step.right_query.dag)

            for resolution in step.config.dependencies:
                if resolution not in self.nodes:
                    raise ValueError(f"Step {resolution} not added to DAG")
            self.graph[step.name] = [parent for parent in step.config.parents]
        else:
            self.graph[step.name] = []

        self.nodes[step.name] = step

    @classmethod
    def list_all(cls) -> list[CollectionName]:
        """List available DAG names on the server."""
        return _handler.list_collections()

    @property
    def run(self) -> RunID:
        """Return run ID if available, else error."""
        if self._run:
            return self._run

        raise RuntimeError(
            "The DAG has not been connected to the server."
            "Start a new run or load a default one."
        )

    @run.setter
    def run(self, run_id: RunID) -> None:
        """Set run ID manually."""
        self._run = run_id

    @property
    def final_steps(self) -> list[Source | Model]:
        """Returns all apex nodes in the DAG.

        Returns:
            List of all apex nodes (nodes with no incoming edges).
            Returns empty list if DAG is empty.
        """
        if not self.nodes:
            return []

        # Find all nodes that appear as dependencies
        all_dependencies = set()
        for deps in self.graph.values():
            all_dependencies.update(deps)

        # Apex nodes are those that don't appear as anyone's dependency
        apex_node_names = [node for node in self.graph if node not in all_dependencies]
        return [self.nodes[name] for name in apex_node_names]

    @property
    def final_step(self) -> Source | Model:
        """Returns the root node in the DAG.

        Returns:
            The root node in the DAG

        Raises:
            ValueError: If the DAG does not have exactly one final step
        """
        steps = self.final_steps

        if len(steps) == 0:
            raise ValueError("No root node found, DAG might contain cycles")
        elif len(steps) > 1:
            raise ValueError("Some models or sources are unreachable")
        else:
            return steps[0]

    def source(self, *args: Any, **kwargs: Any) -> Source:
        """Create Source and add it to the DAG."""
        source = Source(*args, **kwargs, dag=self)
        self._add_step(source)

        return source

    def model(self, *args: Any, **kwargs: Any) -> Model:
        """Create Model and add it to the DAG."""
        model = Model(*args, **kwargs, dag=self)
        self._add_step(model)

        return model

    def add_resolution(
        self,
        name: ResolutionName,
        resolution: Resolution,
    ) -> None:
        """Convert a resolution from the server to a Source or Model and add to DAG."""
        if resolution.resolution_type == ResolutionType.SOURCE:
            self.source(
                location=Location.from_config(resolution.config.location_config),
                name=SourceResolutionName(name),
                extract_transform=resolution.config.extract_transform,
                key_field=resolution.config.key_field,
                index_fields=resolution.config.index_fields,
                description=resolution.description,
                infer_types=False,
                validate_etl=False,
            )
        elif resolution.resolution_type == ResolutionType.MODEL:
            self.model(
                name=ModelResolutionName(name),
                description=resolution.description,
                model_class=resolution.config.model_class,
                model_settings=json.loads(resolution.config.model_settings),
                left_query=Query.from_config(resolution.config.left_query, dag=self),
                right_query=Query.from_config(resolution.config.right_query, dag=self)
                if resolution.config.right_query
                else None,
                truth=threshold_int_to_float(resolution.truth),
            )
        else:
            raise ValueError(f"Unknown resolution type {resolution.resolution_type}")

    def get_source(self, name: ResolutionName) -> Source:
        """Get a source by name from the DAG.

        Args:
            name: The name of the source to retrieve.

        Returns:
            The Source object.

        Raises:
            ValueError: If the name doesn't exist in the DAG or isn't a Source.
        """
        if name not in self.nodes:
            raise ValueError(f"Source '{name}' not found in DAG")

        node = self.nodes[name]
        if not isinstance(node, Source):
            raise ValueError(
                f"Node '{name}' is not a Source, it's a {type(node).__name__}"
            )

        return node

    def get_model(self, name: ResolutionName) -> Model:
        """Get a model by name from the DAG.

        Args:
            name: The name of the model to retrieve.

        Returns:
            The Model object.

        Raises:
            ValueError: If the name doesn't exist in the DAG or isn't a Model.
        """
        if name not in self.nodes:
            raise ValueError(f"Model '{name}' not found in DAG")

        node = self.nodes[name]
        if not isinstance(node, Model):
            raise ValueError(
                f"Node '{name}' is not a Model, it's a {type(node).__name__}"
            )

        return node

    def query(self, *args: Any, **kwargs: Any) -> Query:
        """Create Query object."""
        return Query(*args, **kwargs, dag=self)

    def draw(self, status: DAGExecutionStatus | None = None) -> str:
        """Create a string representation of the DAG as a tree structure.

        If `status` is provided, it will show the status of each node.
        The status indicators are:

        * ✅ Done
        * 🔄 Working
        * ⏸️ Awaiting
        * ⏭️ Skipped

        Args:
            status: Object describing the status of each node.

        Returns:
            String representation of the DAG with status indicators.
        """
        # Handle empty DAG
        if not self.nodes:
            return "Empty DAG"

        apex_nodes = self.final_steps
        if not apex_nodes:
            return "No apex nodes found (possible cycle in DAG)"

        def _get_node_indicator(name: str) -> str:
            """Determine the status indicator for a node."""
            if name not in status:
                return "⏸️"
            elif status[name] == DAGNodeExecutionStatus.SKIPPED:
                return "⏭️"
            elif status[name] == DAGNodeExecutionStatus.DOING:
                return "🔄"
            elif status[name] == DAGNodeExecutionStatus.DONE:
                return "✅"

        # Header with collection and run info
        head_collection: str = f"Collection: {self.name}"
        head_run: str = f"└── Run: {self._run or '⛓️‍💥 Disconnected'}"

        result: list[str] = [head_collection, head_run, ""]
        visited = set()

        def format_children(node: str, prefix: str = "") -> None:
            """Recursively format the children of a node."""
            children = []
            # Get all outgoing edges from this node
            for target in self.graph.get(node, []):
                if target not in visited:
                    children.append(target)
                    visited.add(target)

            # Format each child
            for i, child in enumerate(children):
                is_last = i == len(children) - 1

                # Add status indicator if status is provided
                if status is not None:
                    indicator = _get_node_indicator(child)
                    child_display = f"{indicator} {child}"
                else:
                    child_display = child

                if is_last:
                    result.append(f"{prefix}└── {child_display}")
                    format_children(child, prefix + "    ")
                else:
                    result.append(f"{prefix}├── {child_display}")
                    format_children(child, prefix + "│   ")

        # Draw each apex node
        for i, apex_node in enumerate(apex_nodes):
            root_name = apex_node.name

            if i > 0:
                result.append("")  # Blank line between disconnected components

            visited.add(root_name)

            if status is not None:
                indicator = _get_node_indicator(root_name)
                result.append(f"{indicator} {root_name}")
            else:
                result.append(root_name)

            format_children(root_name)

        return "\n".join(result)

    def __str__(self) -> str:
        """Return string representation of the DAG."""
        return self.draw()

    def new_run(self) -> Self:
        """Start a new run."""
        try:
            collection = _handler.get_collection(self.name)
        except MatchboxCollectionNotFoundError:
            _handler.create_collection(self.name)
            collection = _handler.get_collection(self.name)

        # Delete non-default runs
        for run in collection.runs:
            if run != collection.default_run:
                _handler.delete_run(collection=self.name, run_id=run, certain=True)

        # Start a new run
        self.run = _handler.create_run(collection=self.name).run_id

        return self

    def set_client(self, client: Any) -> Self:  # noqa: ANN401
        """Assign a client to all sources at once."""
        for node in self.nodes.values():
            if isinstance(node, Source):
                node.location.set_client(client)

        return self

    def _load_run(self, run_id: RunID) -> Self:
        """Attach the specified run ID to the current DAG.

        Topologically sorts using Kahn's algorithm.

        Args:
            run_id: The ID of the run to attach
        """
        run: Run = _handler.get_run(collection=self.name, run_id=run_id)
        self.run: RunID = run_id

        resolutions: dict[ResolutionName, Resolution] = run.resolutions

        # Build dependency graph and track added
        added: set[ResolutionName] = set()
        deps_graph: dict[ResolutionName, set[ResolutionName]] = {
            name: {dep for dep in res.config.dependencies}
            for name, res in resolutions.items()
        }

        # Kahn's algorithm: iteratively add resolutions whose dependencies are satisfied
        while len(added) < len(resolutions):
            # Find resolutions that can be added (all dependencies satisfied)
            ready = [
                name
                for name in resolutions
                if name not in added and deps_graph[name].issubset(added)
            ]

            if not ready:
                # No progress possible - circular dependency or missing dependency
                missing = set(resolutions.keys()) - added
                raise RuntimeError(
                    f"Cannot resolve dependencies. Remaining resolutions: {missing}. "
                    "This may indicate circular dependencies or references to "
                    "non-existent resolutions."
                )

            # Add all ready resolutions (order within this batch doesn't matter)
            for name in ready:
                self.add_resolution(name=name, resolution=resolutions[name])
                added.add(name)

        return self

    def load_default(self) -> Self:
        """Attach to default run in this collection, loading all DAG nodes."""
        collection: Collection = _handler.get_collection(self.name)

        if not collection.default_run:
            raise RuntimeError("No default run set.")

        return self._load_run(collection.default_run)

    def load_pending(self) -> Self:
        """Attach to the pending run in this collection, loading all DAG nodes.

        Pending is defined as the last non-default run.
        """
        collection: Collection = _handler.get_collection(self.name)

        pending_runs: list[RunID] = [
            run_id for run_id in collection.runs if run_id != collection.default_run
        ]

        if not pending_runs:
            raise RuntimeError("No pending runs available.")

        return self._load_run(pending_runs[0])

    def run_and_sync(
        self,
        start: str | None = None,
        finish: str | None = None,
        low_memory: bool = False,
        batch_size: int | None = None,
        profile: bool = False,
    ) -> None:
        """Run entire DAG and send results to server.

        Args:
            start: Name of first node to run
            finish: Name of last node to run
            low_memory: Whether to delete data for each node after it is run
            batch_size: The size used for internal batching. Overrides environment
                variable if set.
            profile: whether to log to INFO level the memory usage
        """
        if batch_size is None:
            batch_size = settings.batch_size

        # Determine order of execution steps
        root_nodes = self.final_steps

        def depth_first(node: str, sequence: list) -> None:
            sequence.append(node)
            for neighbour in self.graph[node]:
                if neighbour not in sequence:
                    depth_first(neighbour, sequence)

        inverse_sequence = []
        for root_node in root_nodes:
            depth_first(root_node.name, inverse_sequence)
        sequence = list(reversed(inverse_sequence))

        # Identify skipped nodes
        skipped_nodes = []
        if start:
            try:
                start_index = sequence.index(start)
                skipped_nodes = sequence[:start_index]
            except ValueError as e:
                raise ValueError(f"Step {start} not in DAG") from e
        else:
            start_index = 0

        # Determine end index
        if finish:
            try:
                end_index = sequence.index(finish) + 1
                skipped_nodes.extend(sequence[end_index:])
            except ValueError as e:
                raise ValueError(f"Step {finish} not in DAG") from e
        else:
            end_index = len(sequence)

        sequence = sequence[start_index:end_index]

        status = {
            step_name: DAGNodeExecutionStatus.SKIPPED for step_name in skipped_nodes
        }

        for step_name in sequence:
            node = self.nodes[step_name]
            status[step_name] = DAGNodeExecutionStatus.DOING
            logger.info("\n" + self.draw(status=status))
            try:
                if isinstance(node, Source):
                    node.run(batch_size=batch_size)
                else:
                    node.run()
                node.sync()
                if profile:
                    log_mem_usage()
            except Exception as e:
                logger.error(f"❌ {node.name} failed: {e}")
                raise e
            status[step_name] = DAGNodeExecutionStatus.DONE

            if low_memory:
                node.clear_data()
                logger.info("Cleared node data")
                if profile:
                    log_mem_usage()
        logger.info("\n" + self.draw(status=status))

    def set_default(self) -> None:
        """Set the current run as the default for the collection.

        Makes it immutable, then moves the default pointer to it.
        """
        # Trigger error if there isn't a single root
        _ = self.final_step

        # tries to get apex, error if it doesn't exist
        _handler.set_run_mutable(collection=self.name, run_id=self.run, mutable=False)
        _handler.set_run_default(collection=self.name, run_id=self.run, default=True)

    def lookup_key(
        self,
        from_source: str,
        to_sources: list[str],
        key: str,
        threshold: float | None = None,
    ) -> dict[str, list[str]]:
        """Matches IDs against the selected backend.

        Args:
            from_source: Name of source the provided key belongs to
            to_sources: Names of sources to find keys in
            key: The value to match from the source. Usually a primary key
            threshold (optional): The threshold to use for creating clusters.
                If None, uses the resolutions' default threshold
                If a float, uses that threshold for the specified resolution, and the
                resolution's cached thresholds for its ancestors

        Returns:
            Dictionary mapping source names to list of keys within that source.

        Examples:
            ```python
            dag.lookup_key(
                from_source="companies_house",
                to_sources=[
                    "datahub_companies",
                    "hmrc_exporters",
                ]
                key="8534735",
            )
            ```
        """
        if threshold:
            if not isinstance(threshold, float):
                raise ValueError("If passed, threshold must be a float")
            threshold = threshold_float_to_int(threshold)
        matches = _handler.match(
            targets=[
                ResolutionPath(name=target, collection=self.name, run=self.run)
                for target in to_sources
            ],
            source=ResolutionPath(name=from_source, collection=self.name, run=self.run),
            key=key,
            resolution=self.final_step.resolution_path,
            threshold=threshold,
        )

        to_sources_results = {m.target.name: list(m.target_id) for m in matches}
        # If no matches, _handler will raise
        return {from_source: list(matches[0].source_id), **to_sources_results}

    @profile_time(kwarg="node")
    def resolve(
        self,
        node: ResolutionName | None = None,
        source_filter: list[str] | None = None,
        location_names: list[str] | None = None,
        threshold: float | None = None,
    ) -> ResolvedMatches:
        """Returns ResolvedMatches, optionally filtering.

        Args:
            node: Name of source or model to resolve within DAG.
                If not provided, will look for an apex.
            source_filter: An optional list of source resolution names to filter by.
            location_names: An optional list of location names to filter by.
            threshold (optional): The threshold to use for creating clusters.
                If None, uses the resolutions' default threshold
                If a float, uses that threshold for the specified resolution, and the
                resolution's cached thresholds for its ancestors
        """
        if threshold:
            if not isinstance(threshold, float):
                raise ValueError("If passed, threshold must be a float")
            threshold = threshold_float_to_int(threshold)
        point_of_truth = self.nodes[node] if node else self.final_step

        available_sources = {
            node_name: self.get_source(node_name)
            for node_name in point_of_truth.sources
        }

        filtered_source_names = list(available_sources.keys())

        if source_filter:
            filtered_source_names = [
                s for s in filtered_source_names if s in source_filter
            ]

        if location_names:
            filtered_source_names = [
                s
                for s in filtered_source_names
                if available_sources[s].config.location_config.name in location_names
            ]

        if not filtered_source_names:
            raise MatchboxResolutionNotFoundError("No compatible source was found")

        resolved_sources: list[Source] = []
        query_results: list[pl.DataFrame] = []
        for source_name in filtered_source_names:
            resolved_sources.append(available_sources[source_name])
            query_results.append(
                pl.from_arrow(
                    _handler.query(
                        source=available_sources[source_name].resolution_path,
                        resolution=point_of_truth.resolution_path,
                        return_leaf_id=True,
                        threshold=threshold,
                    )
                )
            )

        return ResolvedMatches(sources=resolved_sources, query_results=query_results)



"""Objects representing the results of running a model client-side."""

from collections.abc import Hashable
from typing import TYPE_CHECKING, Any, ParamSpec, Self, TypeVar

import polars as pl
from pydantic import ConfigDict

from matchbox.client.sources import Source
from matchbox.common.arrow import SCHEMA_RESULTS
from matchbox.common.hash import IntMap
from matchbox.common.logging import logger
from matchbox.common.transform import DisjointSet, to_clusters

if TYPE_CHECKING:
    from matchbox.client.dags import DAG
else:
    DAG = Any

T = TypeVar("T", bound=Hashable)
P = ParamSpec("P")
R = TypeVar("R")


class ModelResults:
    """Results of a model run.

    Contains:

    * The probabilities of each pair being a match
    * (Optional) The clusters of connected components at each threshold
    * (Optional) The leaf_id mapping to trace results back to source clusters

    Allows users to easily interrogate the outputs of models, explore decisions on
    choosing thresholds for clustering, and upload the results to Matchbox.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    probabilities: pl.DataFrame
    _clusters: pl.DataFrame | None = None

    def __init__(
        self,
        probabilities: pl.DataFrame,
        left_root_leaf: pl.DataFrame | None = None,
        right_root_leaf: pl.DataFrame | None = None,
    ) -> None:
        """Initialises and validates results.

        Args:
            probabilities: dataframe with SCHEMA_RESULTS
            left_root_leaf: optional dataframe with columns: id, leaf_id
            right_root_leaf: optional dataframe with columns: id, leaf_id
        """
        self.left_root_leaf = None
        self.right_root_leaf = None

        if left_root_leaf is not None:
            self.left_root_leaf = left_root_leaf
        if right_root_leaf is not None:
            self.right_root_leaf = right_root_leaf

        if not isinstance(probabilities, pl.DataFrame):
            raise ValueError(f"Expected a polars DataFrame, got {type(probabilities)}.")

        expected_fields = set(SCHEMA_RESULTS.names)
        if set(probabilities.columns) != expected_fields:
            raise ValueError(
                f"Expected {expected_fields}.\nFound {set(probabilities.column_names)}."
            )

        # Handle empty tables
        if probabilities.height == 0:
            probabilities = pl.DataFrame(schema=pl.Schema(SCHEMA_RESULTS))

        unique_probabilities = (
            probabilities.with_columns(
                pl.concat_list(
                    [pl.col("left_id").cast(pl.Utf8), pl.col("right_id").cast(pl.Utf8)]
                )
                .list.sort()
                .list.join("_")
                .alias("sorted_ids")
            )
            .sort(
                "probability", descending=True
            )  # sort so largest probability comes first
            .unique(
                subset=["sorted_ids"], keep="first"
            )  # keep first occurrence after sorting
        ).drop("sorted_ids")
        if len(probabilities) != len(unique_probabilities):
            logger.warning(
                "Duplicate pairs! Keeping only pairs with highest probability."
            )

        # Process probability field if it contains floating-point or decimal values
        probability_type = unique_probabilities["probability"].dtype
        if probability_type.is_float() or probability_type.is_decimal():
            probability_uint8 = pl.Series(
                unique_probabilities.select(
                    pl.col("probability").mul(100).round(0).cast(pl.UInt8)
                )
            )

            # Check max value only if the table is not empty
            max_prob = probability_uint8.max()
            if max_prob is not None and max_prob > 100:
                p_max = max_prob
                p_min = probability_uint8.min()
                raise ValueError(f"Probability range misconfigured: [{p_min}, {p_max}]")

            unique_probabilities = unique_probabilities.replace_column(
                unique_probabilities.get_column_index("probability"), probability_uint8
            )

        # Need schema in format recognised by polars
        self.probabilities = unique_probabilities.cast(pl.Schema(SCHEMA_RESULTS))

    @property
    def clusters(self) -> pl.DataFrame:
        """Retrieve new clusters implied by these results."""
        if self._clusters is None:
            im = IntMap()
            self._clusters = to_clusters(
                results=self.probabilities, dtype=pl.Int64, hash_func=im.index
            )
        return self._clusters

    def root_leaf(self) -> pl.DataFrame:
        """Returns all roots and leaves implied by these results."""
        if self.left_root_leaf is None:
            raise RuntimeError(
                "This Results object wasn't instantiated for validation features."
            )

        parents_root_leaf = self.left_root_leaf.select(["id", "leaf_id"])
        if self.right_root_leaf is not None:
            parents_root_leaf = pl.concat(
                [
                    parents_root_leaf,
                    self.right_root_leaf.select(["id", "leaf_id"]),
                ]
            )

        # Go from parent-child (where child could be the root of another model)
        # to root-leaf, where leaf is a source cluster ID
        root_leaf_res = (
            self.clusters.rename({"parent": "root_id"})
            .join(parents_root_leaf, left_on="child", right_on="id")
            .select(["root_id", "leaf_id"])
            .unique()
        )

        # Generate root-leaf for those input rows that weren't merged by this model
        unmerged_ids_rows = (
            parents_root_leaf.select("id", "leaf_id")
            .join(
                self.clusters.select("child"),
                left_on="id",
                right_on="child",
                how="anti",
            )
            .rename({"id": "root_id"})
            .select(["root_id", "leaf_id"])
            .unique()
        )

        return pl.concat([root_leaf_res, unmerged_ids_rows])


class ResolvedMatches:
    """Matches according to resolution."""

    def __init__(
        self, sources: list[Source], query_results: list[pl.DataFrame]
    ) -> None:
        """Initialise resolved data.

        Args:
            sources: List of Source objects
            query_results: List of tables with SCHEMA_QUERY_WITH_LEAVES
        """
        self.sources = sources
        self.query_results = query_results

        if not len(sources):
            raise ValueError("At least 1 source must be resolved.")

        if len(sources) != len(query_results):
            raise ValueError("Mismatched length of sources and query results.")

    @classmethod
    def from_dump(cls, cluster_key_map: pl.DataFrame, dag: DAG) -> Self:
        """Initialise ResolvedMatches from concatenated dataframe representation."""
        partitioned = cluster_key_map.partition_by("source")
        sources = [dag.get_source(p["source"][0]) for p in partitioned]
        query_results = [p.drop("source") for p in partitioned]

        return ResolvedMatches(sources=sources, query_results=query_results)

    def as_lookup(self) -> pl.DataFrame:
        """Return lookup across matchbox ID and source keys."""
        lookup = (
            self.query_results[0]
            .rename({"key": self.sources[0].config.qualified_key(self.sources[0].name)})
            .drop("leaf_id")
        )

        if len(self.sources) > 1:
            for source, source_results in zip(
                self.sources[1:], self.query_results[1:], strict=True
            ):
                lookup = (
                    lookup.join(source_results, on="id", how="full")
                    .with_columns(pl.coalesce(["id", "id_right"]).alias("id"))
                    .drop(["id_right"])
                )

                lookup = lookup.rename(
                    {"key": source.config.qualified_key(source.name)}
                ).drop("leaf_id")

        return lookup

    def as_dump(self) -> pl.DataFrame:
        """Return mapping across root, leaf, source and keys."""
        concat_dfs = []
        for source, query_res in zip(self.sources, self.query_results, strict=True):
            source_col = pl.lit(source.name).alias("source")
            df_with_source = query_res.with_columns([source_col])
            concat_dfs.append(df_with_source)
        return pl.DataFrame(pl.concat(concat_dfs))

    def as_leaf_sets(self) -> list[list[int]]:
        """Return grouping of lead IDs."""
        cluster_key_map = self.as_dump()
        groups = cluster_key_map.group_by("id").agg("leaf_id")["leaf_id"].to_list()
        return [sorted(set(g)) for g in groups]

    def view_cluster(self, cluster_id: int, merge_fields: bool = False) -> pl.DataFrame:
        """Return source data for all records in cluster.

        Args:
            cluster_id: ID of root cluster to view
            merge_fields: whether to remove source qualifier when concatenating rows.
                Only applies to index fields - key fields are not affected.
        """
        cluster_rows = []
        key_cols = []
        for source, query_res in zip(self.sources, self.query_results, strict=True):
            # For each source, get rows for selected cluster
            source_keys = query_res.filter(pl.col("id") == cluster_id)["key"].to_list()
            if not source_keys:
                continue

            key_cols.append(source.qualified_key)
            # Determine column names of output dataframe
            rename_keys = {source.key_field.name: source.qualified_key}
            if not merge_fields:
                rename_index_fields = {
                    field.name: source.qualify_field(field.name)
                    for field in source.index_fields
                }
                rename_dict = {**rename_keys, **rename_index_fields}
            else:
                rename_dict = rename_keys

            # Fetch data for this source
            source_data = pl.concat(
                source.fetch(keys=source_keys, qualify_names=False)
            ).rename(rename_dict)

            cluster_rows.append(source_data)

        # Coerce fields to their common super-type
        if not cluster_rows:
            raise KeyError(f"Cluster {cluster_id} not available")
        source_concat = pl.concat(cluster_rows, how="diagonal_relaxed")
        # Re-order columns to have keys at the beginning
        remaining_cols = [col for col in source_concat.columns if col not in key_cols]

        return source_concat.select(*key_cols, *remaining_cols)

    def merge(self, other: Self) -> Self:
        """Combine two instances of resolved matches by merging clusters.

        All cluster IDs will be replaced with negative integers and lose their
        association with cluster IDs on the backend.
        """
        if other.sources != self.sources:
            raise ValueError("Cannot merge resolved matches for different sources")

        djs = DisjointSet[int]()

        # Use disjoint sets to find new clusters
        for resolved_instance in [self, other]:
            unioned_root_leaf = pl.concat(
                [
                    result.select(["id", "leaf_id"])
                    for result in resolved_instance.query_results
                ]
            )
            components = (
                unioned_root_leaf.group_by("id")
                .agg(pl.col("leaf_id"))["leaf_id"]
                .to_list()
            )
            # Create component by connecting one leaf to all others within component
            for leaf_set in components:
                djs.add(leaf_set[0])
                for other_leaf in leaf_set[1:]:
                    djs.union(leaf_set[0], other_leaf)

        # Turn new clusters into dataframe
        new_components = []
        for i, component in enumerate(djs.get_components(), start=1):
            cluster_id = -i
            for leaf_id in component:
                new_components.append({"id": cluster_id, "leaf_id": leaf_id})

        new_components_df = pl.DataFrame(new_components)

        # Generate new combined query_results based on new clusters
        new_query_results = []
        for self_result, other_result in zip(
            self.query_results, other.query_results, strict=True
        ):
            unioned_leaf_key = pl.concat(
                [
                    self_result.select("leaf_id", "key"),
                    other_result.select("leaf_id", "key"),
                ]
            ).unique()
            source_query_results = unioned_leaf_key.join(
                new_components_df, on="leaf_id"
            )
            new_query_results.append(source_query_results)

        return ResolvedMatches(sources=self.sources, query_results=new_query_results)



"""Client-side helpers for retrieving and preparing evaluation samples."""

import polars as pl
from pydantic import BaseModel
from sqlalchemy.exc import OperationalError

from matchbox.client import _handler
from matchbox.client.dags import DAG
from matchbox.client.results import ModelResults
from matchbox.common.dtos import ModelResolutionName, ModelResolutionPath, SourceConfig
from matchbox.common.eval import Judgement, precision_recall
from matchbox.common.exceptions import MatchboxSourceTableError


class EvaluationFieldMetadata(BaseModel):
    """Metadata for a field in evaluation."""

    display_name: str
    source_columns: list[str]


class EvaluationItem(BaseModel):
    """A cluster ready for evaluation.

    The records dataframe contains the leaf IDs and the qualified index fields
    associated with it. For example:

    | leaf_id | src_a_first | src_a_last | src_b_first | src_b_last |
    |---------|-------------|------------|-------------|------------|
    | 1       | Thomas      | Bayes      |             |            |
    | 2       | Tommy       | B          |             |            |
    | 12      |             |            | Tom         | Bayes      |

    The fields attribute allows any evaluation system to map between a display
    version of the source columns, and the actual columns contained in the
    records dataframe. For example:

    ```text
    {
        "display_name": "first",
        "source_columns": "src_a_first", "src_b_first"
    }
    ```
    """

    model_config = {"arbitrary_types_allowed": True}

    leaves: list[int]
    records: pl.DataFrame
    fields: list[EvaluationFieldMetadata]

    def get_unique_record_groups(self) -> list[list[int]]:
        """Group identical records by leaf ID.

        Returns:
            List of groups, where each group is a list of leaf IDs
            that have identical values across all data fields.
            Example: [[1, 3], [2], [4, 5, 6]] means records 1 & 3 are identical.
        """
        # Get all data column names (not "leaf")
        # Flatten the source_columns lists from all fields
        data_cols = [col for field in self.fields for col in field.source_columns]

        # Group by all data columns to find duplicates
        grouped = self.records.group_by(data_cols, maintain_order=True).agg(
            pl.col("leaf")
        )

        # Extract list of leaf ID lists
        return [group for group in grouped["leaf"]]


def create_judgement(
    item: EvaluationItem,
    assignments: dict[int, str],
    user_name: str,
    tag: str | None = None,
) -> Judgement:
    """Convert item assignments to Judgement - no default group assignment.

    Args:
        item: evaluation item
        assignments: column assignments (group_idx -> group_letter)
        user_name: user name for the judgement
        tag: string by which to tag the judgement

    Returns:
        Judgement with endorsed groups based on assignments
    """
    groups: dict[str, list[int]] = {}
    unique_record_groups = item.get_unique_record_groups()

    for col_idx, group in assignments.items():
        leaf_ids = unique_record_groups[col_idx]
        groups.setdefault(group, []).extend(leaf_ids)

    endorsed = [sorted(set(leaf_ids)) for leaf_ids in groups.values()]
    return Judgement(user_name=user_name, shown=item.leaves, endorsed=endorsed, tag=tag)


def create_evaluation_item(
    df: pl.DataFrame, source_configs: list[tuple[str, SourceConfig]], leaves: list[int]
) -> EvaluationItem:
    """Create EvaluationItem with structured metadata."""
    # Get all data columns (exclude metadata columns)
    data_cols = [c for c in df.columns if c not in ["root", "leaf", "key"]]

    # Build mapping of field_name -> list of qualified column names
    field_to_columns: dict[str, list[str]] = {}

    for source_id, config in source_configs:
        for field in config.index_fields:
            # Build qualified column name for this source+field
            column_name = config.qualify_field(source_id, field.name)

            # Only add if this column exists in DataFrame
            if column_name in data_cols:
                # Add to list for this field name
                if field.name not in field_to_columns:
                    field_to_columns[field.name] = []
                field_to_columns[field.name].append(column_name)

    # Create EvaluationFieldMetadata objects (one per unique field name)
    fields: list[EvaluationFieldMetadata] = []
    for field_name, source_columns in field_to_columns.items():
        fields.append(
            EvaluationFieldMetadata(
                display_name=field_name, source_columns=source_columns
            )
        )

    # Keep ALL data columns in records
    records = df.select(["leaf"] + data_cols)

    return EvaluationItem(leaves=leaves, records=records, fields=fields)


def _read_sample_file(sample_file: str, n: int) -> pl.DataFrame:
    resolved_matches_dump = pl.read_parquet(sample_file)
    clusters = resolved_matches_dump["id"].unique()
    select_clusters = clusters.sample(min(n, len(clusters)), shuffle=True).to_list()
    select_rows = resolved_matches_dump.filter(pl.col("id").is_in(select_clusters))
    return select_rows.rename({"id": "root", "leaf_id": "leaf"})


def _get_samples_from_server(
    dag: DAG, user_name: str, n: int, resolution: ModelResolutionName | None = None
) -> pl.DataFrame:
    if resolution:
        resolution_path: ModelResolutionPath = dag.get_model(resolution).resolution_path
    else:
        resolution_path: ModelResolutionPath = dag.final_step.resolution_path
    return pl.from_arrow(
        _handler.sample_for_eval(n=n, resolution=resolution_path, user_name=user_name)
    )


def get_samples(
    n: int,
    dag: DAG,
    user_name: str,
    resolution: ModelResolutionName | None = None,
    sample_file: str | None = None,
) -> dict[int, EvaluationItem]:
    """Retrieve samples enriched with source data as EvaluationItems.

    Args:
        n: Number of clusters to sample
        dag: DAG for which to retrieve samples
        user_name: Name of the user requesting the samples
        resolution: The optional resolution from which to sample. If not set, the final
            step in the DAG is used. If sample_file is set, resolution is ignored
        sample_file: path to parquet file output by ResolvedMatches. If specified,
            won't sample from server, ignoring the user_name and resolution arguments

    Returns:
        Dictionary of cluster ID to EvaluationItems describing the cluster

    Raises:
        MatchboxSourceTableError: If a source cannot be queried from a location using
            provided or default clients.
    """
    if sample_file:
        samples = _read_sample_file(sample_file=sample_file, n=n)
    else:
        samples = _get_samples_from_server(
            dag=dag, user_name=user_name, n=n, resolution=resolution
        )

    if not len(samples):
        return {}

    results_by_source: list[pl.DataFrame] = []
    source_configs: list[tuple[str, SourceConfig]] = []

    for source_resolution in samples["source"].unique():
        try:
            source = dag.get_source(source_resolution)
        except ValueError as exc:  # pragma: no cover - defensive path
            raise MatchboxSourceTableError(
                f"Source '{source_resolution}' not found in DAG. "
                "Ensure DAG was loaded with all resolutions."
            ) from exc

        source_configs.append((source_resolution, source.config))

        samples_by_source = samples.filter(pl.col("source") == source_resolution)
        keys_by_source = samples_by_source["key"].to_list()

        try:
            source_data = pl.concat(
                source.fetch(batch_size=10_000, qualify_names=True, keys=keys_by_source)
            )
        except OperationalError as exc:
            raise MatchboxSourceTableError(
                f"Could not query source '{source_resolution}' from warehouse. "
                "Check warehouse connection and ensure source table exists."
            ) from exc

        samples_and_source = samples_by_source.join(
            source_data, left_on="key", right_on=source.qualified_key
        )
        desired_columns = ["root", "leaf", "key"] + source.qualified_index_fields
        results_by_source.append(samples_and_source[desired_columns])

    if not results_by_source:
        return {}

    all_results: pl.DataFrame = pl.concat(results_by_source, how="diagonal")

    results_by_root: dict[int, EvaluationItem] = {}
    for root in all_results["root"].unique():
        cluster_df = all_results.filter(pl.col("root") == root).drop("root")
        leaves = cluster_df.select("leaf").to_series().unique().to_list()
        evaluation_item = create_evaluation_item(cluster_df, source_configs, leaves)
        results_by_root[root] = evaluation_item

    return results_by_root


class EvalData:
    """Object which caches evaluation data to measure model performance."""

    def __init__(self, tag: str | None = None) -> None:
        """Download judgement and expansion data used to compute evaluation metrics."""
        self.tag = tag
        self.judgements, self.expansion = _handler.download_eval_data(tag)

    def precision_recall(
        self, results: ModelResults, threshold: float
    ) -> tuple[float, float]:
        """Compute precision and recall for a given Results object."""
        if not len(results.clusters):
            raise ValueError("No clusters suggested by these results.")

        threshold = int(threshold * 100)
        root_leaf = results.root_leaf().rename({"root_id": "root", "leaf_id": "leaf"})
        values = precision_recall([root_leaf], self.judgements, self.expansion)[0]
        return values[0], values[1]



"""CLI tool for EdDSA key pair generation and JWT token creation."""

import json
import sys
import time
from base64 import urlsafe_b64encode
from typing import Annotated

import typer
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from rich import print, print_json

app = typer.Typer()

EXPIRY_AFTER_X_HOURS = 24
DEFAULT_PRIVATE_KEY = "private_key.pem"
DEFAULT_PUBLIC_KEY = "public_key.pem"


def b64encode_nopadding(to_encode: bytes) -> bytes:
    """B64 encode without padding."""
    return urlsafe_b64encode(to_encode).rstrip(b"=")


def generate_EdDSA_key_pair() -> tuple[bytes, bytes]:
    """Generate private and public key pair."""
    private_key = Ed25519PrivateKey.generate()

    unencrypted_pem_private_key = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    pem_public_key = private_key.public_key().public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    return unencrypted_pem_private_key, pem_public_key


def generate_json_web_token(
    private_key_bytes: bytes,
    sub: str,
    api_root: str = "https://api.example.com",
    expiry_hours: int = EXPIRY_AFTER_X_HOURS,
    email: str | None = None,
) -> str:
    """Generate JWT with private key bytes."""
    private_key = load_pem_private_key(private_key_bytes, password=None)

    header = {
        "typ": "JWT",
        "alg": "EdDSA",
        "crv": "Ed25519",
    }
    payload = {
        "sub": sub,
        "exp": int(time.time() + 60 * 60 * expiry_hours),
        "authorised_hosts": api_root,
    }

    if email is not None:
        payload["email"] = email

    to_sign = (
        b64encode_nopadding(json.dumps(header).encode("utf-8"))
        + b"."
        + b64encode_nopadding(json.dumps(payload).encode("utf-8"))
    )
    signature = b64encode_nopadding(private_key.sign(to_sign))
    token = (to_sign + b"." + signature).decode()
    return token


# CLI commands
@app.command()
def keygen() -> None:
    """Generate an EdDSA key pair and output as JSON.

    Pipe to a .json to save.
    """
    unencrypted_pem_private_key, pem_public_key = generate_EdDSA_key_pair()

    output = {
        "private_key": unencrypted_pem_private_key.decode("utf-8"),
        "public_key": pem_public_key.decode("utf-8"),
    }

    print_json(json.dumps(output))


@app.command()
def jwt(
    sub: Annotated[
        str,
        typer.Option(
            "--sub",
            "-s",
            help="Subject claim for the JWT (should be a UUID string)",
        ),
    ],
    email: Annotated[
        str | None,
        typer.Option(
            "--email",
            "-m",
            help="Optional email address to include in the JWT",
        ),
    ] = None,
    api_root: Annotated[
        str,
        typer.Option(
            "--api-root",
            "-a",
            help="Authorised API host",
        ),
    ] = "https://api.example.com",
    expiry_hours: Annotated[
        int,
        typer.Option(
            "--expiry",
            "-e",
            help="Token expiry in hours",
        ),
    ] = EXPIRY_AFTER_X_HOURS,
    private_key: Annotated[
        typer.FileText | None,
        typer.Option(
            "--key",
            "-k",
            help="Private key file (or use stdin)",
        ),
    ] = None,
) -> None:
    """Generate JWT from private key.
    
    Use with keygen like:

    uv run test/scripts/authorisation.py keygen > keys.json
    
    cat keys.json | \
    jq -r .private_key | \
    uv run test/scripts/authorisation.py jwt \
        --sub a1b2c3d4-e5f6-7890-abcd-ef1234567890 \
        --email user@example.com \
        --api-root http://api.example.com/ \
        --expiry 2
    
    Or with a key file:
    
    uv run test/scripts/authorisation.py jwt \
        --key private_key.pem \
        --sub f9e8d7c6-b5a4-3210-9876-543210fedcba \
        --email user@example.com
    """

    if private_key:
        private_key_pem = private_key.read()
    elif not sys.stdin.isatty():
        private_key_pem = sys.stdin.read()
    else:
        typer.echo(
            "Error: No input piped. Provide a private key via stdin or --key.", err=True
        )
        raise typer.Exit(code=1)

    token = generate_json_web_token(
        private_key_bytes=private_key_pem.encode(),
        sub=sub,
        api_root=api_root,
        expiry_hours=expiry_hours,
        email=email,
    )

    print(token)


if __name__ == "__main__":
    app()



"""Factories for generating sources and linked source testkits for testing."""

import warnings
from collections.abc import Callable
from functools import cache, wraps
from itertools import product
from math import prod
from typing import Any, ParamSpec, Self, TypeVar

import pandas as pd
import polars as pl
import pyarrow as pa
from adbc_driver_manager.dbapi import Connection as AdbcConnection
from faker import Faker
from pydantic import BaseModel, ConfigDict, Field, field_validator
from sqlalchemy import Engine, create_engine
from sqlglot import cast, select
from sqlglot.expressions import column

from matchbox.client.dags import DAG
from matchbox.client.locations import RelationalDBLocation
from matchbox.client.sources import Source
from matchbox.common.arrow import SCHEMA_INDEX, SCHEMA_QUERY
from matchbox.common.datatypes import DataTypes
from matchbox.common.dtos import (
    SourceConfig,
    SourceField,
    SourceResolutionName,
    SourceResolutionPath,
)
from matchbox.common.factories.entities import (
    ClusterEntity,
    EntityReference,
    FeatureConfig,
    SourceEntity,
    SuffixRule,
    diff_results,
    generate_entities,
    probabilities_to_results_entities,
)
from matchbox.common.hash import hash_values

P = ParamSpec("P")
R = TypeVar("R")


def make_features_hashable(func: Callable[P, R]) -> Callable[P, R]:
    """Decorator to allow configuring source_factory with dicts.

    This retains the hashability of FeatureConfig while still making it simple
    to use the factory without special objects.
    """

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        # Handle features in first positional arg
        if args and args[0] is not None:
            if isinstance(args[0][0], dict):
                args = (tuple(FeatureConfig(**d) for d in args[0]),) + args[1:]
            else:
                args = (tuple(args[0]),) + args[1:]

        # Handle features in kwargs
        if "features" in kwargs and kwargs["features"] is not None:
            if isinstance(kwargs["features"][0], dict):
                kwargs["features"] = tuple(
                    FeatureConfig(**d) for d in kwargs["features"]
                )
            else:
                kwargs["features"] = tuple(kwargs["features"])

        return func(*args, **kwargs)

    return wrapper


class SourceTestkitParameters(BaseModel):
    """Configuration for generating a source."""

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    features: tuple[FeatureConfig, ...] = Field(default_factory=tuple)
    name: str
    engine: Engine | AdbcConnection = Field(default=create_engine("sqlite:///:memory:"))
    n_true_entities: int | None = Field(default=None)
    repetition: int = Field(default=0)


class SourceTestkit(BaseModel):
    """A testkit of data and metadata for a SourceConfig."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    source: Source = Field(
        description="The Source object containing config and convenience methods."
    )
    features: tuple[FeatureConfig, ...] | None = Field(
        description=(
            "The features used to generate the data. "
            "If None, the source data was not generated, but set manually."
        ),
        default=None,
    )
    data: pa.Table = Field(
        description="The generated data, corresponding to the output of queries."
    )
    data_hashes: pa.Table = Field(description="A PyArrow table of hashes for the data.")
    entities: tuple[ClusterEntity, ...] = Field(
        description="ClusterEntities that were generated from the source."
    )

    @field_validator("data")
    @classmethod
    def cast_table(cls, value: pa.Table) -> pa.Table:
        """Ensure that the data matches the query schema."""
        for col in SCHEMA_QUERY.names:
            cast_col = value[col].cast(SCHEMA_QUERY.field(col).type)
            value = value.set_column(value.schema.get_field_index(col), col, cast_col)

        return value

    @property
    def name(self) -> str:
        """Return the source name."""
        return self.source.name

    @property
    def resolution_path(self) -> SourceResolutionPath:
        """Returns the source resolution path."""
        return self.source.resolution_path

    @property
    def source_config(self) -> SourceConfig:
        """Return the SourceConfig from the source."""
        return self.source.config

    def fake_run(self) -> Self:
        """Set source hashes before source is run."""
        self.source.hashes = self.data_hashes

        return self

    def into_dag(self) -> dict:
        """Turn source into kwargs for `dag.source()`, detaching from original DAG."""
        return {
            "location": self.source.location,
            "name": self.source.name,
            "extract_transform": self.source.config.extract_transform,
            "key_field": self.source.config.key_field,
            "index_fields": self.source.config.index_fields,
            "description": self.source.description,
        }

    def write_to_location(self, set_client: Any | None = None) -> Self:  # noqa: ANN401
        """Write the data to the SourceConfig's location.

        Args:
            set_client: client to replace existing source client
        """
        if set_client:
            self.source.location.set_client(set_client)

        pl.from_arrow(self.data).write_database(
            table_name=self.source.name,
            connection=self.source.location.client,
            if_table_exists="replace",
        )

        return self


class LinkedSourcesTestkit(BaseModel):
    """Container for multiple related SourceConfig testkits with entity tracking."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    dag: DAG
    true_entities: set[SourceEntity] = Field(default_factory=set)
    sources: dict[SourceResolutionName, SourceTestkit]

    def find_entities(
        self,
        min_appearances: dict[str, int] | None = None,
        max_appearances: dict[str, int] | None = None,
    ) -> list[SourceEntity]:
        """Find entities matching appearance criteria."""
        result = list(self.true_entities)

        def _meets_criteria(
            entity: SourceEntity, criteria: dict[str, int], compare: Callable
        ) -> bool:
            return all(
                compare(len(entity.get_keys(src)), count)
                for src, count in criteria.items()
            )

        if min_appearances:
            result = [
                entity
                for entity in result
                if _meets_criteria(entity, min_appearances, lambda x, y: x >= y)
            ]

        if max_appearances:
            result = [
                entity
                for entity in result
                if _meets_criteria(entity, max_appearances, lambda x, y: x <= y)
            ]

        return result

    def true_entity_subset(self, *sources: SourceResolutionName) -> list[ClusterEntity]:
        """Return a subset of true entities that appear in the given sources."""
        cluster_entities = [
            entity.to_cluster_entity(*sources) for entity in self.true_entities
        ]
        return [entity for entity in cluster_entities if entity is not None]

    def diff_results(
        self,
        probabilities: pl.DataFrame,
        sources: list[SourceResolutionName],
        left_clusters: tuple[ClusterEntity, ...],
        right_clusters: tuple[ClusterEntity, ...] | None = None,
        threshold: int | float = 0,
    ) -> tuple[bool, dict]:
        """Diff a results of probabilities with the true SourceEntities.

        Args:
            probabilities: Probabilities table to diff
            sources: Subset of the LinkedSourcesTestkit.sources that represents
                the true sources to compare against
            left_clusters: ClusterEntity objects from the object used as an input
                to the process that produced the probabilities table. Should
                be a SourceTestkit.entities or ModelTestkit.entities.
            right_clusters: ClusterEntity objects from the object used as an input
                to the process that produced the probabilities table. Should
                be a SourceTestkit.entities or ModelTestkit.entities.
            threshold: Threshold for considering a match true

        Returns:
            A tuple of whether the results are identical, and a report dictionary.
                See [`diff_results()`][matchbox.common.factories.entities.diff_results]
                for the report format.
        """
        return diff_results(
            expected=self.true_entity_subset(*sources),
            actual=probabilities_to_results_entities(
                probabilities=probabilities,
                left_clusters=left_clusters,
                right_clusters=right_clusters,
                threshold=threshold,
            ),
        )

    def write_to_location(self) -> Self:
        """Write the data to the SourceConfig's location."""
        for source_testkit in self.sources.values():
            source_testkit.write_to_location()

        return self


def generate_rows(
    generator: Faker,
    selected_entities: tuple[SourceEntity, ...],
    features: tuple[FeatureConfig, ...],
    repetition: int,
) -> tuple[
    dict[str, list], dict[int, list[str]], dict[int, list[str]], dict[int, bytes]
]:
    """Generate raw data rows with unique keys and shared IDs.

    This function generates rows of data plus maps between three types of identifiers:

        1. `id`: Is matchbox's unique identifier for each row, shared across rows with
            identical feature values
        2. `key`: Is the source's unique identifier for the row. It's like a primary key
            in a database, but not guaranteed to be unique across different entities
        3. `entity`: Is the identifier of the SourceEntity that generated the row.
            This identifies the true linked data in the factory system.

    This function will therefore return:

        * raw_data: A dictionary of column arrays for DataFrame creation
        * entity_keys: A dictionary that maps which keys belong to each source entity
        * id_keys: A dictionary that maps which keys share the same row content,
            with the same `id`
        * id_hashes: A dictionary that maps `id`s to hash values for each unique
            row content

    The key insight:

        * entity_* groups by "who generated this row"
        * id_* groups by "what content does this row have"

    Example with two entities generating data:

    | id | key | company_name |
    |----|-----|--------------|
    | 1  | a   | alpha co     |
    | 2  | b   | alpha ltd    |
    | 1  | c   | alpha co     |  # Same content as row 'a'
    | 2  | d   | alpha ltd    |  # Same content as row 'b'
    | 3  | e   | beta co      |
    | 4  | f   | beta ltd     |
    | 3  | g   | beta co      |  # Same content as row 'e'
    | 4  | h   | beta ltd     |  # Same content as row 'f'

    What does this table look like as raw data?

    ```python
    raw_data = {
        "id": [1, 2, 1, 2, 3, 4, 3, 4],
        "key": ["a", "b", "c", "d", "e", "f", "g", "h"],
        "company_name": [
            "alpha co",
            "alpha ltd",
            "alpha co",
            "alpha ltd",
            "beta co",
            "beta ltd",
            "beta co",
            "beta ltd",
        ],
    }
    ```

    Which keys came from each source entity?

    ```python
    entity_keys = {
        1: ["a", "b", "c", "d"],  # All keys entity 1 produced
        2: ["e", "f", "g", "h"],  # All keys entity 2 produced
    }
    ```

    Which keys have identical content?

    ```python
    id_keys = {
        1: ["a", "c"],  # Both have "alpha co" content
        2: ["b", "d"],  # Both have "alpha ltd" content
        3: ["e", "g"],  # Both have "beta co" content
        4: ["f", "h"],  # Both have "beta ltd" content
    }
    id_hashes = {
        1: b"hash1",  # Hash of "alpha co"
        2: b"hash2",  # Hash of "alpha ltd"
        3: b"hash3",  # Hash of "beta co"
        4: b"hash4",  # Hash of "beta ltd"
    }
    ```
    """
    raw_data = {"key": [], "id": []}
    for feature in features:
        raw_data[feature.name] = []

    # Track entity locations and row identities
    entity_keys = {entity.id: [] for entity in selected_entities}
    id_keys = {}
    id_hashes = {}
    value_to_id = {}

    def add_row(entity_id: int, values: tuple) -> None:
        """Add a row of data, handling IDs and keys."""
        key = str(generator.uuid4())
        entity_keys[entity_id].append(key)
        row_hash = hash_values(*(str(v) for v in values))

        if values not in value_to_id:
            mb_id = generator.random_number(digits=16)
            value_to_id[values] = mb_id
            id_keys[mb_id] = []
            id_hashes[mb_id] = row_hash

        row_id = value_to_id[values]
        id_keys[row_id].append(key)

        raw_data["key"].append(key)
        raw_data["id"].append(row_id)
        for feature, value in zip(features, values, strict=True):
            raw_data[feature.name].append(value)

    for entity in selected_entities:
        # For each feature, collect all possible values
        possible_values = []
        for feature in features:
            base = entity.base_values[feature.name]

            for rule in feature.variations:
                if not isinstance(base, rule.type):
                    raise TypeError(
                        f"{rule.__class__.__name__} requires {rule.type} "
                        f"but {feature.name} generates {type(base)}"
                    )

            variations = []
            # Apply all variations as long as they change the value
            for v in (rule.apply(base) for rule in feature.variations):
                if v != base:
                    variations.append(v)

            values = variations if feature.drop_base else variations + [base]
            possible_values.append(values or [base])

        if (num_variations := prod(len(values) for values in possible_values)) > 100:
            raise RuntimeError(
                f"Entity {entity.id} would generate {num_variations:,} variations, "
                "which exceeds the maximum of 100. This would cause a Cartesian "
                "explosion."
            )

        for values in product(*possible_values):
            for _ in range(repetition + 1):
                add_row(entity.id, values)

    return raw_data, entity_keys, id_keys, id_hashes


@cache
def generate_source(
    generator: Faker,
    n_true_entities: int,
    features: tuple[FeatureConfig, ...],
    repetition: int,
    seed_entities: tuple[SourceEntity, ...] | None = None,
) -> tuple[pa.Table, pa.Table, dict[int, set[str]], dict[int, set[str]]]:
    """Generate raw data as PyArrow tables with entity tracking.

    Returns:
        - data: PyArrow table with generated data
        - data_hashes: PyArrow table with hash groups
        - entity_keys: SourceEntity ID -> list of keys mapping
        - id_keys: Unique row ID -> list of keys mapping for identical rows
    """
    # Select or generate entities
    if seed_entities is None:
        selected_entities = generate_entities(generator, features, n_true_entities)
    else:
        selected_entities = generator.random_elements(
            elements=seed_entities,
            unique=True,
            length=min(n_true_entities, len(seed_entities)),
        )

    # Generate initial data
    raw_data, entity_keys, id_keys, id_hashes = generate_rows(
        generator=generator,
        selected_entities=selected_entities,
        features=features,
        repetition=repetition,
    )

    # Create DataFrame
    df = pd.DataFrame(raw_data)

    # Create hash groups
    keys = []
    hashes = []
    for row_id, group_keys in id_keys.items():
        keys.append(list(group_keys))
        hashes.append(id_hashes[row_id])

    data_hashes = pa.Table.from_pydict(
        {
            "keys": keys,
            "hash": hashes,
        },
        schema=SCHEMA_INDEX,
    )

    # Update variation counts
    for entity in selected_entities:
        if entity.id in entity_keys:
            # Count unique row IDs this entity appears in
            entity_rows = df[df["key"].isin(entity_keys[entity.id])]
            entity.total_unique_variations = len(set(entity_rows["id"]))

    return (
        pa.Table.from_pandas(df, preserve_index=False),
        data_hashes,
        entity_keys,
        id_keys,
    )


@make_features_hashable
@cache
def source_factory(
    features: list[FeatureConfig] | list[dict] | None = None,
    name: SourceResolutionName | None = None,
    location_name: str = "dbname",
    dag: DAG | None = None,
    engine: Engine | None = None,
    n_true_entities: int = 10,
    repetition: int = 0,
    seed: int = 42,
) -> SourceTestkit:
    """Generate a complete source testkit from configured features.

    SourceConfigs created with the factory system can only use a RelationalDBLocation,
    and the data at that location will be stored in a single table.

    Args:
        features: List of FeatureConfig objects or dictionaries to use for generating
            the source data. If None, defaults to a set of common features.
        name: Name of the source. If None, a unique name is generated. This will be
            used as the name of the table in the RelationalDBLocation, but also in
            the SourceResolutionName for the source.
        location_name: Name of the location for the source.
        dag: DAG containing the source.
        engine: SQLAlchemy engine to use for the source's RelationalDBLocation. If
            None, an in-memory SQLite engine is created.
        n_true_entities: Number of true entities to generate. Defaults to 10.
        repetition: Number of times to repeat the generated data. Defaults to 0.
        seed: Random seed for reproducibility. Defaults to 42.
    """
    generator = Faker()
    generator.seed_instance(seed)

    if features is None:
        features = (
            FeatureConfig(
                name="company_name",
                base_generator="company",
            ),
            FeatureConfig(
                name="crn",
                base_generator="bothify",
                parameters=(("text", "???-###-???-###"),),
            ),
        )

    if name is None:
        name = generator.unique.word()

    if engine is None:
        engine = create_engine("sqlite:///:memory:")

    if dag is None:
        dag = DAG("collection")
        dag.run = 1

    # Generate base entities
    base_entities = generate_entities(
        generator=generator,
        features=features,
        n=n_true_entities,
    )

    # Generate data using the base entities
    data, data_hashes, entity_keys, row_keys = generate_source(
        generator=generator,
        n_true_entities=n_true_entities,
        features=features,
        repetition=repetition,
        seed_entities=base_entities,
    )

    # Create source entities with references
    source_entities = []
    for entity in base_entities:
        keys = entity_keys.get(entity.id, [])
        if keys:
            entity.add_source_reference(name, keys)
            source_entities.append(entity)

    # Create ClusterEntity objects from row_keys
    results_entities = [
        ClusterEntity(
            id=row_id,
            keys=EntityReference({name: frozenset(keys)}),
        )
        for row_id, keys in row_keys.items()
    ]

    # Create fields
    key_field = SourceField(name="key", type=DataTypes.STRING)
    index_fields = tuple(
        SourceField(name=feature.name, type=feature.datatype) for feature in features
    )

    # Create source config
    source = Source(
        dag=dag,
        location=RelationalDBLocation(name=location_name).set_client(engine),
        name=name,
        description=f"Generated source for {name}",
        extract_transform=select(
            cast(column(key_field.name), "string").as_(key_field.name),
            *[column(field.name) for field in index_fields],
        )
        .from_(name)
        .sql(),
        key_field=key_field,
        index_fields=index_fields,
    )

    return SourceTestkit(
        source=source,
        features=features,
        data=data,
        data_hashes=data_hashes,
        entities=tuple(sorted(results_entities)),
    )


def source_from_tuple(
    data_tuple: tuple[dict[str, Any], ...],
    data_keys: tuple[Any],
    name: str | None = None,
    location_name: str = "dbname",
    dag: DAG | None = None,
    engine: Engine | None = None,
    seed: int = 42,
) -> SourceTestkit:
    """Generate a complete source testkit from dummy data."""
    generator = Faker()
    generator.seed_instance(seed)

    if name is None:
        name = generator.unique.word()

    if engine is None:
        engine = create_engine("sqlite:///:memory:")

    if dag is None:
        dag = DAG("collection")
        dag.run = 1

    base_entities = tuple(SourceEntity(base_values=row) for row in data_tuple)

    # Create source entities with references
    source_entities: list[SourceEntity] = []
    for entity, key in zip(base_entities, data_keys, strict=True):
        entity.add_source_reference(name, [key])
        source_entities.append(entity)
    entity_ids = {entity.id for entity in source_entities}

    # Create ClusterEntity objects from row_keys
    results_entities = [
        ClusterEntity(
            id=entity_id,
            keys=EntityReference({name: frozenset([key])}),
        )
        for key, entity_id in zip(data_keys, entity_ids, strict=True)
    ]

    # Create fields
    key_field = SourceField(name="key", type=DataTypes.STRING)
    index_fields = tuple(
        SourceField(name=k, type=DataTypes.from_pytype(type(v)))
        for k, v in data_tuple[0].items()
    )

    # Create source config
    source = Source(
        dag=dag,
        location=RelationalDBLocation(name=location_name).set_client(engine),
        name=name,
        description=f"Generated source for {name}",
        extract_transform=select(
            cast(column(key_field.name), "string").as_(key_field.name),
            *[column(field.name) for field in index_fields],
        )
        .from_(name)
        .sql(),
        key_field=key_field,
        index_fields=index_fields,
    )

    hashes = [hash_values(*row) for row in data_tuple]

    data_hashes = pa.Table.from_pydict(
        {
            # Assumes that string conversion will be the same as the SQL warehouse's
            "keys": [str(dkey) for dkey in data_keys],
            "hash": hashes,
        },
        schema=SCHEMA_INDEX,
    )

    raw_data = pa.Table.from_pylist(list(data_tuple))
    raw_keys = pa.array(data_keys)

    data = raw_data.append_column("id", [entity_ids]).append_column("key", raw_keys)

    return SourceTestkit(
        source=source,
        data=data,
        data_hashes=data_hashes,
        entities=tuple(sorted(results_entities)),
    )


@cache
def linked_sources_factory(
    source_parameters: tuple[SourceTestkitParameters, ...] | None = None,
    n_true_entities: int | None = None,
    engine: Engine | None = None,
    dag: DAG | None = None,
    seed: int = 42,
) -> LinkedSourcesTestkit:
    """Generate a set of linked sources with tracked entities.

    Args:
        source_parameters: Optional tuple of source testkit parameters
        n_true_entities: Optional number of true entities to generate. If provided,
            overrides any n_true_entities in source configs. If not provided, each
            SourceTestkitParameters must specify its own n_true_entities.
        engine: Optional SQLAlchemy engine to use for all sources. If provided,
            overrides any engine in source configs.
        dag: DAG containing sources
        seed: Random seed for reproducibility
    """
    generator = Faker()
    generator.seed_instance(seed)

    if dag is None:
        dag = DAG("collection")
        dag.run = 1

    default_engine = create_engine("sqlite:///:memory:")

    if source_parameters is None:
        # Use factory parameter or default for default configs
        n_true_entities = n_true_entities if n_true_entities is not None else 10

        features = {
            "company_name": FeatureConfig(
                name="company_name",
                base_generator="company",
            ),
            "crn": FeatureConfig(
                name="crn",
                base_generator="bothify",
                parameters=(("text", "???-###-???-###"),),
            ),
            "dh": FeatureConfig(
                name="dh",
                base_generator="numerify",
                parameters=(("text", "########"),),
            ),
            "cdms": FeatureConfig(
                name="cdms",
                base_generator="numerify",
                parameters=(("text", "ORG-########"),),
            ),
            "address": FeatureConfig(
                name="address",
                base_generator="address",
            ),
        }

        source_parameters = (
            SourceTestkitParameters(
                name="crn",
                engine=engine or default_engine,
                features=(
                    features["company_name"].add_variations(
                        SuffixRule(suffix=" Limited"),
                        SuffixRule(suffix=" UK"),
                        SuffixRule(suffix=" Company"),
                    ),
                    features["crn"],
                ),
                n_true_entities=n_true_entities,
                repetition=0,
            ),
            SourceTestkitParameters(
                name="dh",
                engine=engine or default_engine,
                features=(
                    features["company_name"],
                    features["dh"],
                ),
                n_true_entities=n_true_entities // 2,
                repetition=0,
            ),
            SourceTestkitParameters(
                name="cdms",
                engine=engine or default_engine,
                features=(
                    features["crn"],
                    features["cdms"],
                ),
                n_true_entities=n_true_entities,
                repetition=1,
            ),
        )
    else:
        if n_true_entities is not None:
            # Factory parameter provided - warn if configs have values set
            config_entities = [
                parameters.n_true_entities for parameters in source_parameters
            ]
            if any(n is not None for n in config_entities):
                warnings.warn(
                    "Both source testkit configs and linked_sources_factory specify "
                    "n_true_entities. The factory parameter will be used.",
                    UserWarning,
                    stacklevel=2,
                )
            # Override all configs with factory parameter
            source_parameters = tuple(
                parameters.model_copy(
                    update={
                        "engine": engine or parameters.engine,
                        "n_true_entities": n_true_entities,
                    }
                )
                for parameters in source_parameters
            )
        else:
            # No factory parameter - check all configs have n_true_entities set
            missing_counts = [
                parameters.name
                for parameters in source_parameters
                if parameters.n_true_entities is None
            ]
            if missing_counts:
                raise ValueError(
                    "n_true_entities not set for sources: "
                    f"{', '.join(missing_counts)}. When factory n_true_entities is "
                    "not provided, all configs must specify it."
                )

    # Collect all unique features
    all_features = set()
    for parameters in source_parameters:
        all_features.update(parameters.features)
    all_features = tuple(sorted(all_features, key=lambda f: f.name))

    # Find maximum number of entities needed across all sources
    max_entities = max(parameters.n_true_entities for parameters in source_parameters)

    # Generate all possible entities
    all_entities = generate_entities(
        generator=generator, features=all_features, n=max_entities
    )

    # Initialize LinkedSourcesTestkit
    true_entity_lookup = {entity.id: entity for entity in all_entities}
    linked = LinkedSourcesTestkit(
        dag=dag,
        true_entities=all_entities,
        sources={},
    )

    # Generate sources
    for parameters in source_parameters:
        # Generate source data using seed entities
        data, data_hashes, entity_keys, row_keys = generate_source(
            generator=generator,
            features=tuple(parameters.features),
            n_true_entities=parameters.n_true_entities,
            repetition=parameters.repetition,
            seed_entities=all_entities,
        )

        # Create ClusterEntity objects from row_keys
        results_entities = [
            ClusterEntity(
                id=row_id,
                keys=EntityReference({parameters.name: frozenset(keys)}),
            )
            for row_id, keys in row_keys.items()
        ]

        # Create fields
        key_field = SourceField(name="key", type=DataTypes.STRING)
        index_fields = tuple(
            SourceField(name=feature.name, type=feature.datatype)
            for feature in parameters.features
        )

        # Create source config
        source = Source(
            dag=dag,
            location=RelationalDBLocation(str(parameters.name)).set_client(
                parameters.engine
            ),
            name=parameters.name,
            description=f"Generated source for {parameters.name}",
            extract_transform=select(
                cast(column(key_field.name), "string").as_(key_field.name),
                *[column(field.name) for field in index_fields],
            )
            .from_(parameters.name)
            .sql(),
            key_field=key_field,
            index_fields=index_fields,
        )

        # Add source to linked.sources
        linked.sources[parameters.name] = SourceTestkit(
            source=source,
            features=tuple(parameters.features),
            data=data,
            data_hashes=data_hashes,
            entities=tuple(sorted(results_entities)),
        )

        # Update entities with source references
        for entity_id, keys in entity_keys.items():
            entity = true_entity_lookup[entity_id]
            entity.add_source_reference(parameters.name, keys)

    return linked



"""API routes for the Matchbox server."""

import uuid
from collections.abc import Awaitable, Callable
from importlib.metadata import version
from pathlib import Path
from typing import Annotated

from fastapi import (
    Depends,
    FastAPI,
    HTTPException,
    Query,
    Request,
    Response,
)
from fastapi.encoders import jsonable_encoder
from fastapi.openapi.docs import (
    get_swagger_ui_html,
    get_swagger_ui_oauth2_redirect_html,
)
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.responses import HTMLResponse

from matchbox.common.arrow import table_to_buffer
from matchbox.common.dtos import (
    BackendCountableType,
    CollectionName,
    CountResult,
    CRUDOperation,
    ErrorResponse,
    Match,
    ModelResolutionName,
    OKMessage,
    ResolutionPath,
    ResourceOperationStatus,
    RunID,
    SourceResolutionName,
    SourceResolutionPath,
)
from matchbox.common.exceptions import MatchboxHttpException
from matchbox.common.logging import logger
from matchbox.server.api.dependencies import (
    BackendDependency,
    ParquetResponse,
    authorisation_dependencies,
    lifespan,
)
from matchbox.server.api.routers import auth, collection, eval

app = FastAPI(
    title="matchbox API",
    version=version("matchbox_db"),
    lifespan=lifespan,
    docs_url=None,
    redoc_url=None,
)
app.include_router(collection.router)
app.include_router(eval.router)
app.include_router(auth.router)

static_dir = Path(__file__).parent / "static"
app.mount("/static", StaticFiles(directory=static_dir), name="static")


@app.exception_handler(StarletteHTTPException)
async def http_exception_handler(
    request: Request, exc: StarletteHTTPException
) -> JSONResponse:
    """Overwrite the default JSON schema for an `HTTPException`."""
    return JSONResponse(
        content=jsonable_encoder(exc.detail), status_code=exc.status_code
    )


@app.exception_handler(Exception)
async def exception_handler(request: Request, exc: Exception) -> JSONResponse:
    """Unified handler for all exceptions."""
    if isinstance(exc, MatchboxHttpException):
        error_response = ErrorResponse(
            exception_type=type(exc).__name__,
            message=str(exc),
            details=exc.to_details(),
        )
        return JSONResponse(
            status_code=exc.http_status,
            content=jsonable_encoder(error_response.model_dump()),
        )

    # All other exceptions get logged and return generic error
    error_id = str(uuid.uuid4())
    logger.exception("Unhandled exception [%s]: %s", error_id, exc)

    error_response = ErrorResponse(
        exception_type="MatchboxServerError",
        message=f"An internal server error occurred. Reference: {error_id}",
        details=None,
    )
    return JSONResponse(
        status_code=500,
        content=jsonable_encoder(error_response.model_dump()),
    )


@app.middleware("http")
async def add_security_headers(
    request: Request, call_next: Callable[[Request], Awaitable[Response]]
) -> Response:
    """Improve security by adding headers to all responses."""
    response: Response = await call_next(request)
    response.headers["Cache-control"] = "no-store, no-cache"
    response.headers["Content-Security-Policy"] = (
        # Restrict by default
        "default-src 'none'; frame-ancestors 'none'; form-action 'none';"
        # Load Swagger CSS, favicon and openapi.json
        "style-src 'self'; img-src 'self' data:; connect-src 'self'; "
        # Load Swagger JS, hard-coding the expected file hash
        "script-src 'self' 'sha256-QOOQu4W1oxGqd2nbXbxiA1Di6OHQOLQD+o+G9oWL8YY='"
    )
    response.headers["Strict-Transport-Security"] = (
        "max-age=31536000; includeSubDomains"
    )
    response.headers["X-Frame-Options"] = "DENY"
    response.headers["X-Content-Type-Options"] = "nosniff"
    return response


@app.get("/health")
async def healthcheck() -> OKMessage:
    """Perform a health check and return the status."""
    return OKMessage()


# Retrieval


@app.get(
    "/query",
    responses={404: {"model": ErrorResponse}},
)
def query(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    source: SourceResolutionName,
    return_leaf_id: bool,
    resolution: ModelResolutionName | None = None,
    threshold: int | None = None,
    limit: int | None = None,
) -> ParquetResponse:
    """Query Matchbox for matches based on a source resolution name."""
    res = backend.query(
        source=SourceResolutionPath(
            collection=collection,
            run=run_id,
            name=source,
        ),
        point_of_truth=ResolutionPath(
            collection=collection, run=run_id, name=resolution
        )
        if resolution
        else None,
        threshold=threshold,
        return_leaf_id=return_leaf_id,
        limit=limit,
    )

    buffer = table_to_buffer(res)
    return ParquetResponse(buffer.getvalue())


@app.get(
    "/match",
    responses={404: {"model": ErrorResponse}},
)
def match(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    targets: Annotated[list[SourceResolutionName], Query()],
    source: SourceResolutionName,
    key: str,
    resolution: ModelResolutionName,
    threshold: int | None = None,
) -> list[Match]:
    """Match a source key against a list of target source resolutions."""
    targets = [
        SourceResolutionPath(collection=collection, run=run_id, name=t) for t in targets
    ]
    return backend.match(
        key=key,
        source=SourceResolutionPath(
            collection=collection,
            run=run_id,
            name=source,
        ),
        targets=targets,
        point_of_truth=ResolutionPath(
            collection=collection,
            run=run_id,
            name=resolution,
        ),
        threshold=threshold,
    )


# Admin


@app.get("/database/count")
def count_backend_items(
    backend: BackendDependency,
    entity: BackendCountableType | None = None,
) -> CountResult:
    """Count the number of various entities in the backend."""

    def get_count(e: BackendCountableType) -> int:
        return getattr(backend, str(e)).count()

    if entity is not None:
        return CountResult(entities={str(entity): get_count(entity)})
    else:
        res = {str(e): get_count(e) for e in BackendCountableType}
        return CountResult(entities=res)


@app.delete(
    "/database/orphans",
    responses={
        500: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
    },
    dependencies=[Depends(authorisation_dependencies)],
)
def delete_orphans(backend: BackendDependency) -> ResourceOperationStatus:
    """Delete orphans."""
    try:
        orphans = backend.delete_orphans()
        return ResourceOperationStatus(
            success=True,
            target="Database",
            operation=CRUDOperation.DELETE,
            details=f"Deleted {orphans} orphans.",
        )
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ResourceOperationStatus(
                success=False,
                target="Database",
                operation=CRUDOperation.DELETE,
                details=str(e),
            ),
        ) from e


@app.delete(
    "/database",
    responses={
        409: {"model": ErrorResponse},
        500: {"model": ErrorResponse},
    },
    dependencies=[Depends(authorisation_dependencies)],
)
def clear_database(
    backend: BackendDependency,
    certain: Annotated[
        bool,
        Query(
            description=(
                "Confirm deletion of all data in the database whilst retaining tables"
            )
        ),
    ] = False,
) -> ResourceOperationStatus:
    """Delete all data from the backend whilst retaining tables."""
    backend.clear(certain=certain)
    return ResourceOperationStatus(
        success=True, target="Database", operation=CRUDOperation.DELETE
    )


# Swagger UI


@app.get("/docs", include_in_schema=False)
async def custom_swagger_ui_html() -> HTMLResponse:
    """Get locally hosted docs."""
    return get_swagger_ui_html(
        openapi_url=app.openapi_url,
        title=app.title + " - Swagger UI",
        oauth2_redirect_url=app.swagger_ui_oauth2_redirect_url,
        swagger_js_url="/static/swagger-ui-bundle.js",
        swagger_css_url="/static/swagger-ui.css",
        swagger_favicon_url="/static/favicon.png",
    )


@app.get(app.swagger_ui_oauth2_redirect_url, include_in_schema=False)
async def swagger_ui_redirect() -> HTMLResponse:
    """Helper for OAuth2."""
    return get_swagger_ui_oauth2_redirect_html()



"""Collections PostgreSQL mixin for Matchbox server."""

from psycopg.errors import LockNotAvailable
from pyarrow import Table
from sqlalchemy import delete, select, update

from matchbox.common.db import sql_to_df
from matchbox.common.dtos import (
    Collection,
    CollectionName,
    ModelResolutionPath,
    Resolution,
    ResolutionPath,
    ResolutionType,
    Run,
    RunID,
    SourceResolutionPath,
    UploadStage,
)
from matchbox.common.dtos import ModelConfig as CommonModelConfig
from matchbox.common.dtos import SourceConfig as CommonSourceConfig
from matchbox.common.exceptions import (
    MatchboxCollectionAlreadyExists,
    MatchboxDeletionNotConfirmed,
    MatchboxLockError,
    MatchboxResolutionUpdateError,
    MatchboxRunNotWriteable,
)
from matchbox.common.logging import logger
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.orm import (
    Collections,
    ModelConfigs,
    Resolutions,
    Results,
    Runs,
    SourceConfigs,
    SourceFields,
)
from matchbox.server.postgresql.utils.db import compile_sql
from matchbox.server.postgresql.utils.insert import insert_hashes, insert_results


class MatchboxPostgresCollectionsMixin:
    """Collections mixin for the PostgreSQL adapter for Matchbox."""

    # Collection management

    def create_collection(self, name: CollectionName) -> Collection:  # noqa: D102
        with MBDB.get_session() as session:
            if (session.query(Collections).filter_by(name=name).first()) is None:
                new_collection = Collections(name=name)
                session.add(new_collection)
                session.commit()
                return new_collection.to_dto()
            else:
                raise MatchboxCollectionAlreadyExists

    def get_collection(  # noqa: D102
        self, name: CollectionName
    ) -> Collection:
        with MBDB.get_session() as session:
            collection_orm = Collections.from_name(name, session)
            return collection_orm.to_dto()

    def list_collections(self) -> list[CollectionName]:  # noqa: D102
        with MBDB.get_session() as session:
            collections = (
                session.execute(select(Collections.name).order_by(Collections.name))
                .scalars()
                .all()
            )
            return list(collections)

    def delete_collection(self, name: CollectionName, certain: bool) -> None:  # noqa: D102
        with MBDB.get_session() as session:
            collection_orm = Collections.from_name(name, session)

            if not certain:
                version_names = [v.name for v in collection_orm.versions]
                raise MatchboxDeletionNotConfirmed(children=version_names)

            session.execute(
                delete(Collections).where(
                    Collections.collection_id == collection_orm.collection_id
                )
            )
            session.commit()

    # Run management

    def create_run(self, collection: CollectionName) -> Run:  # noqa: D102
        with MBDB.get_session() as session:
            # Can raise MatchboxCollectionNotFoundError
            collection_orm = Collections.from_name(collection, session)

            new_run = Runs(
                collection_id=collection_orm.collection_id,
                is_mutable=True,
                is_default=False,
            )
            session.add(new_run)
            session.commit()

            return new_run.to_dto()

    def set_run_mutable(  # noqa: D102
        self, collection: CollectionName, run_id: RunID, mutable: bool
    ) -> Run:
        with MBDB.get_session() as session:
            run_orm = Runs.from_id(collection, run_id, session)
            run_orm.is_mutable = mutable
            session.commit()

            return run_orm.to_dto()

    def set_run_default(  # noqa: D102
        self, collection: CollectionName, run_id: RunID, default: bool
    ) -> Run:
        with MBDB.get_session() as session:
            run_orm = Runs.from_id(collection, run_id, session)
            if default:
                if run_orm.is_mutable:
                    raise ValueError("Cannot set as default a mutable run")
                # Unset any existing default run for the collection
                session.execute(
                    update(Runs)
                    .where(
                        Runs.collection_id == run_orm.collection_id,
                        Runs.is_default.is_(True),
                    )
                    .values(is_default=False)
                )

            run_orm.is_default = default
            session.commit()

            return run_orm.to_dto()

    def get_run(self, collection: CollectionName, run_id: RunID) -> Run:  # noqa: D102
        with MBDB.get_session() as session:
            run_orm = Runs.from_id(collection, run_id, session)
            return run_orm.to_dto()

    def delete_run(  # noqa: D102
        self, collection: CollectionName, run_id: RunID, certain: bool
    ) -> None:
        with MBDB.get_session() as session:
            run_orm = Runs.from_id(collection, run_id, session)

            if not certain:
                resolution_names = [res.name for res in run_orm.resolutions]
                raise MatchboxDeletionNotConfirmed(children=resolution_names)

            session.execute(delete(Runs).where(Runs.run_id == run_orm.run_id))
            session.commit()

    # Resolution management

    def _check_writeable(self, path: ResolutionPath) -> None:
        run = Runs.from_id(collection=path.collection, run_id=path.run)
        if not run.is_mutable:
            raise MatchboxRunNotWriteable(
                f"Version {path.run} in collection {path.collection} is immutable"
            )

    def create_resolution(  # noqa: D102
        self, resolution: Resolution, path: ResolutionPath
    ) -> None:
        self._check_writeable(path)
        log_prefix = f"Insert {path.name}"
        with MBDB.get_session() as session:
            resolution_orm = Resolutions.from_dto(
                resolution=resolution, path=path, session=session
            )
            session.commit()

            logger.info(
                f"Inserted with ID {resolution_orm.resolution_id}", prefix=log_prefix
            )

    def get_resolution(  # noqa: D102
        self, path: ResolutionPath
    ) -> Resolution:
        with MBDB.get_session() as session:
            resolution = Resolutions.from_path(path=path, session=session)
            return resolution.to_dto()

    def update_resolution(  # noqa: D102
        self, resolution: Resolution, path: ResolutionPath
    ) -> None:
        new_config = resolution.config
        with MBDB.get_session() as session:
            # Get current ORM entry
            old_resolution = Resolutions.from_path(path=path, session=session)
            # Check current ORM entry can be updated
            if old_resolution.fingerprint != resolution.fingerprint:
                raise MatchboxResolutionUpdateError(
                    "Cannot update resolution with non-matching fingerprint."
                )
            # The following condition also protects against change of resolution type:
            # sources must have 0 parents, models must have 1 or more
            if old_resolution.to_dto().config.parents != new_config.parents:
                raise MatchboxResolutionUpdateError(
                    "Cannot change parents of a resolution."
                )

            # Update top-level metadata
            old_resolution.description = resolution.description
            old_resolution.truth = resolution.truth

            # Update config
            if old_resolution.type == "source":
                old_config: SourceConfigs = old_resolution.source_config
                if not isinstance(new_config, CommonSourceConfig):
                    raise ValueError("Config for source resolution expected.")
                old_config.location_name = new_config.location_config.name
                old_config.location_type = str(new_config.location_config.type)
                old_config.extract_transform = new_config.extract_transform

                # Update source fields
                if (
                    old_config.to_dto().key_field != new_config.key_field
                    or old_config.to_dto().index_fields != new_config.index_fields
                ):
                    # If any field differs, delete and start again
                    old_config.fields.clear()
                    session.flush()

                    new_fields = [
                        SourceFields(
                            index=0,
                            name=new_config.key_field.name,
                            is_key=True,
                            type=new_config.key_field.type.value,
                        )
                    ]
                    new_fields.extend(
                        [
                            SourceFields(
                                index=idx + 1,
                                name=field.name,
                                is_key=False,
                                type=field.type.value,
                            )
                            for idx, field in enumerate(new_config.index_fields)
                        ]
                    )

                    old_config.fields = new_fields

            else:
                old_config: ModelConfigs = old_resolution.model_config
                if not isinstance(new_config, CommonModelConfig):
                    raise ValueError("Config for model resolution expected.")
                old_config.model_class = new_config.model_class
                old_config.model_settings = new_config.model_settings
                old_config.left_query = new_config.left_query.model_dump_json()
                old_config.right_query = (
                    None
                    if not new_config.right_query
                    else new_config.right_query.model_dump_json()
                )

            session.commit()

    def delete_resolution(self, path: ResolutionPath, certain: bool) -> None:  # noqa: D102
        self._check_writeable(path)
        with MBDB.get_session() as session:
            resolution = Resolutions.from_path(path=path, session=session)
            if certain:
                delete_stmt = delete(Resolutions).where(
                    Resolutions.resolution_id.in_(
                        [
                            resolution.resolution_id,
                            *(d.resolution_id for d in resolution.descendants),
                        ]
                    )
                )
                session.execute(delete_stmt)
                session.commit()
            else:
                children = [r.name for r in resolution.descendants]
                raise MatchboxDeletionNotConfirmed(children=children)

    # Data insertion

    def lock_resolution_data(self, path: ResolutionPath) -> None:  # noqa: D102
        self._check_writeable(path)
        with MBDB.get_session() as session:
            # Lock resolution so only one client can initiate the upload
            # Will fail if already locked
            try:
                resolution = Resolutions.from_path(
                    path=path, session=session, for_update=True
                )
            except LockNotAvailable as e:
                raise MatchboxLockError("Resolution is locked.") from e

            # Check status
            # Will fail if stage not READY
            if resolution.upload_stage == UploadStage.COMPLETE:
                session.rollback()
                raise MatchboxLockError(
                    "Once set to complete, resolution data stage cannot be changed."
                )
            elif resolution.upload_stage == UploadStage.PROCESSING:
                session.rollback()
                raise MatchboxLockError("Upload already being processed.")

            resolution.upload_stage = UploadStage.PROCESSING
            session.commit()

    def unlock_resolution_data(  # noqa: D102
        self, path: ResolutionPath, complete: bool = False
    ) -> None:
        self._check_writeable(path)
        with MBDB.get_session() as session:
            resolution = Resolutions.from_path(
                path=path, session=session, for_update=True
            )
            if complete:
                resolution.upload_stage = UploadStage.COMPLETE
            else:
                resolution.upload_stage = UploadStage.READY
            session.commit()

    def get_resolution_stage(self, path: ResolutionPath) -> UploadStage:  # noqa: D102
        resolution = Resolutions.from_path(path)
        return UploadStage(resolution.upload_stage)

    def insert_source_data(  # noqa: D102
        self, path: SourceResolutionPath, data_hashes: Table
    ) -> None:
        self._check_writeable(path)
        insert_hashes(
            path=path, data_hashes=data_hashes, batch_size=self.settings.batch_size
        )
        self.unlock_resolution_data(path=path, complete=True)

    def insert_model_data(self, path: ModelResolutionPath, results: Table) -> None:  # noqa: D102
        self._check_writeable(path)
        insert_results(path=path, results=results, batch_size=self.settings.batch_size)
        self.unlock_resolution_data(path=path, complete=True)

    def get_model_data(self, path: ModelResolutionPath) -> Table:  # noqa: D102
        with MBDB.get_session() as session:
            resolution = Resolutions.from_path(
                path=path, res_type=ResolutionType.MODEL, session=session
            )

            results_query = select(
                Results.left_id, Results.right_id, Results.probability
            ).where(Results.resolution_id == resolution.resolution_id)

        with MBDB.get_adbc_connection() as conn:
            stmt: str = compile_sql(results_query)
            return sql_to_df(
                stmt=stmt, connection=conn.dbapi_connection, return_type="arrow"
            )
